#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import numbers
import math
import json
from collections import OrderedDict

import numpy

from titus.signature import LabelData
from titus.datatype import ForwardDeclarationParser
from titus.datatype import AvroUnion
import titus.prettypfa

class Dataset(object):
    """Canonical format for providing a dataset to the tree-builder.

    Constructors are __init__ and fromIterable.
    """

    class Field(object):
        """Represents a field of the dataset; usually created by a Dataset constructor.

        ``Dataset.Field`` objects may be in one of two states: Numpy and
        Python.  The Dataset constructors produce ``Dataset.Fields`` in
        their Numpy representation, which is required by the
        tree-builder.  In the Numpy representation, categorical string
        data are represented as integers from 0 to N-1, where N is the
        number of unique input strings, with each distinct integer
        representing a distinct input string.  Strings and integers
        can be converted through the ``intToStr`` and ``strToInt``
        dictionaries, or by converting the whole array into a Pythonic
        form with the ``toPython`` method.
        """

        def __init__(self, t):
            self.tpe = t
            self.data = []

        def __repr__(self):
            return "<Dataset.Field of type {0} at 0x{1:08x}>".format("float" if self.tpe == numbers.Real else "str", id(self))

        def add(self, v):
            self.data.append(v)

        def toNumpy(self):
            """Changes this field into a Numpy representation *in-place* (destructively replaces the old representation)."""

            if self.tpe == numbers.Real:
                self.data = numpy.array(self.data, dtype=numpy.dtype(float))
            elif self.tpe == str:
                unique = sorted(set(self.data))
                intToStr = dict(enumerate(unique))
                strToInt = dict((x, i) for i, x in enumerate(unique))
                converted = numpy.empty(len(self.data), dtype=numpy.dtype(int))
                for i, x in enumerate(self.data):
                    converted[i] = strToInt[x]
                self.intToStr = intToStr
                self.strToInt = strToInt
                self.data = converted
            return self

        def select(self, selection):
            """Creates a new ``Dataset.Field`` from this one by applying a boolean-valued Numpy array of the same length.

            The new ``Dataset.Field`` is independent of the old one (this is a purely functional method).

            Assumes that the ``Dataset.Field`` is currently in a Numpy representation.

            :type selection: 1-d Numpy array of bool
            :param selection: data points to select
            :rtype: 1-d Numpy array
            :return: subset of the original ``self.data``
            """

            out = Dataset.Field(self.tpe)
            out.data = self.data[selection]
            if hasattr(self, "intToStr"):
                out.intToStr = self.intToStr
                out.strToInt = self.strToInt
            return out

        def toPython(self):
            """Changes this field into a Python representation *in-place* (destructively replaces the old representation)."""

            if self.tpe == numbers.Real:
                self.data = list(self.data)
            elif self.tpe == str:
                converter = numpy.array([x for i, x in sorted(self.intToStr.items())], dtype=numpy.dtype(object))
                self.data = list(converter[self.data])
            return self
        
    def __init__(self, fields, names):
        self.fields = fields
        self.names = names

    @classmethod
    def fromIterable(cls, iterable, limit=None, names=None):
        """Constructor for Dataset that takes a Python iterable (rows) of iterables (columns).

        Each row must have the same number of fields with the same types (``numbers.Real`` or ``basestring``).

        :type iterable: Python iterable
        :param iterable: input dataset
        :type limit: positive integer or ``None``
        :param limit: maximum number of input rows
        :type names: list of strings or ``None``
        :param names: names of the fields; if not provided, names like ``var0``, ``var1``, etc. will be generated.
        :rtype: titus.producer.cart.Dataset
        :return: a dataset
        """

        fields = []
        for lineNumber, line in enumerate(iterable):
            if lineNumber == 0:
                for word in line:
                    if isinstance(word, numbers.Real):
                        fields.append(cls.Field(numbers.Real))
                    elif isinstance(word, str):
                        fields.append(cls.Field(str))
                    else:
                        raise ValueError("record type must be a real number or a string, not {0}".format(type(word)))
                if names is None:
                    formatter = "var{0:0%dd}" % len(str(len(line)))
                    names = [formatter.format(i) for i in range(len(line))]
                else:
                    if len(names) != len(fields):
                        raise ValueError("number of columns in dataset is not the same as the number of names")

            if limit is not None and lineNumber >= limit:
                break

            if len(line) != len(fields):
                raise ValueError("number of columns in dataset is not the same as the first: {0} on line {1}".format(len(line), lineNumber))

            for columnNumber, (word, field) in enumerate(zip(line, fields)):
                if isinstance(word, field.tpe):
                    field.add(word)
                else:
                    raise ValueError("type of column {0} in dataset is not the same as the first: {1} on line {2}".format(columnNumber, type(word), lineNumber))

        return cls([x.toNumpy() for x in fields], names)

    def __len__(self):
        return len(self.fields[0].data)

    def __repr__(self):
        return "<Dataset with {0} fields at 0x{1:08x}>".format(len(self.fields), id(self))

class TreeNode(object):
    """Represents a tree node and applies the CART algorithm to build decision and regression trees.

    The constructors are ``__init__`` and ``fromWholeDataset``.

    Tree-building is initiated by calling ``splitUntil(condition)``, where ``condition(node, depth)`` is a user-supplied function that takes a node (titus.producer.cart.TreeNode) and depth (integer) and returns bool (``True``: continue splitting; ``False``: stop splitting).
    """

    @classmethod
    def fromWholeDataset(cls, wholeDataset, predictandName, maxSubsetSize=None):
        """Constructor for a tree from a dataset that includes the predictand (that which we try to purify in the leaves) as one of its fields.

        :type wholeDataset: titus.producer.cart.Dataset
        :param wholeDataset: dataset including the predictand
        :type predictandName: string
        :param predictandName: name of the predictand, to be taken out of the dataset
        :type maxSubsetSize: positive integer or ``None``
        :param maxSubsetSize: maximum size of subset splits of categorical regressors (approximation for optimization in ``categoricalEntropyGainTerm`` and ``categoricalNVarianceGainTerm``)
        :rtype: titus.producer.cart.TreeNode
        :return: an unsplit tree
        """

        predictandIndex = wholeDataset.names.index(predictandName)
        dataset = Dataset(wholeDataset.fields[:predictandIndex] + wholeDataset.fields[(predictandIndex + 1):],
                          wholeDataset.names[:predictandIndex] + wholeDataset.names[(predictandIndex + 1):])
        predictand = wholeDataset.fields[predictandIndex]
        maxSubsetSize = maxSubsetSize
        return cls(dataset, predictand, maxSubsetSize)

    def __init__(self, dataset, predictand, maxSubsetSize=None):
        """Constructor for a tree from a dataset of regressors (that which we split) and a predictand (that which we try to purify in the leaves).

        :type dataset: titus.producer.cart.Dataset
        :param dataset: dataset of regressors only
        :type predictand: 1-d Numpy array
        :param predictand: predictands in a separate array with the same number of rows as the ``dataset``
        :type maxSubsetSize: positive integer or ``None``
        :param maxSubsetSize: maximum size of subset splits of categorical regressors (approximation for optimization in ``categoricalEntropyGainTerm`` and ``categoricalNVarianceGainTerm``)
        """

        self.dataset = dataset
        self.predictand = predictand
        self.maxSubsetSize = maxSubsetSize

        self.datasetSize = len(self.predictand.data)

        if self.predictand.tpe == numbers.Real:
            try:
                self.predictandUnique = numpy.unique(self.predictand.data)
            except TypeError:
                self.predictandUnique = numpy.unique1d(self.predictand.data)

        elif self.predictand.tpe == str:
            if self.datasetSize > 0:
                self.predictandDistribution = []
                for category in range(len(self.predictand.intToStr)):
                    frac = 1.0 * numpy.sum(self.predictand.data == category) / len(self.predictand.data)
                    self.predictandDistribution.append(frac)
            else:
                self.predictandDistribution = [0.0] * len(self.predictand.intToStr)

        else:
            raise RuntimeError

    def splitComplete(self):
        """Convenience function for building up a tree until each leaf has only one unique value. Calls ``splitUntil``."""

        self.splitUntil(lambda node, depth: True)

    def splitMaxDepth(self, maxDepth):
        """Convenience function for building up trees until each leaf has only one unique value or the depth reaches ``maxDepth``. Calls ``splitUntil``.

        :type maxDepth: positive integer
        :param maxDepth: maximum allowed depth of the tree
        """

        self.splitUntil(lambda node, depth: depth < maxDepth)

    def splitUntil(self, condition, depth=1):
        """Performs a recursive tree-split, calling the user-supplied ``condition(node, depth)`` at each new node.

        If the predictand is numerical (``numbers.Real``), the node has attributes: ``datasetSize``, ``predictandUnique``, ``nTimesVariance``, and ``gain``.

        If the predictand is categorical (``basestring``), the node has attributes: ``datasetSize``, ``predictandDistribution``, ``entropy``, and ``gain``.

        Splits are performed *in-place*, changing this ``TreeNode``.

        :type condition: callable that takes node (titus.producer.cart.TreeNode) and depth (integer) and returns bool (``True``: continue splitting; ``False``: stop splitting).
        :param condition: splitting condition function
        :type depth: positive integer
        :param depth: current depth
        """

        if self.canSplit():
            self.splitOnce()
            if hasattr(self, "passBranch") and hasattr(self, "failBranch") and condition(self, depth):
                self.passBranch.splitUntil(condition, depth + 1)
                self.failBranch.splitUntil(condition, depth + 1)

    def canSplit(self):
        """Returns ``True`` if it is possible to split the predictand; ``False`` otherwise."""

        if self.predictand.tpe == numbers.Real:
            return len(self.predictandUnique) > 1
        elif self.predictand.tpe == str:
            return numpy.count_nonzero(self.predictandDistribution) > 0
        else:
            raise RuntimeError

    def score(self):
        """Returns the best score at this ``TreeNode``, which might or might not be a leaf."""

        if self.predictand.tpe == numbers.Real:
            return numpy.mean(self.predictand.data)
        elif self.predictand.tpe == str:
            return self.predictand.intToStr[numpy.argmax(self.predictandDistribution)]
        else:
            raise RuntimeError
    
    def splitOnce(self):
        """Compute an optimized split in one field, adding two new ``TreeNodes`` below this one.

        If the predictand is numerical (``numbers.Real``), the split minimizes entropy; if categorical (``basestring``), it minimizes n-times-variance.
        """

        # build a regression tree using n-times-variance as the metric to optimize
        if self.predictand.tpe == numbers.Real:
            self.nTimesVariance = len(self.predictand.data) * numpy.var(self.predictand.data)

            self.gain = None
            self.split = None
            self.fieldIndex = None
            self.field = None
            # for each field...
            for fieldIndex, field in enumerate(self.dataset.fields):
                if field.tpe == numbers.Real:
                    gainTerm, split = self.numericalNVarianceGainTerm(field)
                elif field.tpe == str:
                    gainTerm, split = self.categoricalNVarianceGainTerm(field, self.maxSubsetSize)
                else:
                    raise RuntimeError

                # the gainTerm functions don't include this constant (n-times-variance of the unsplit node)
                gainTerm += self.nTimesVariance

                # ... and keep track of the best field, best split
                if self.gain is None or gainTerm > self.gain:
                    self.gain = gainTerm
                    self.split = split
                    self.fieldIndex = fieldIndex
                    self.field = field
            
        # build a classification tree using entropy as the metric to optimize
        elif self.predictand.tpe == str:
            # find the best split by maximizing entropic gain
            self.entropy = 0.0
            for frac in self.predictandDistribution:
                if frac != 0.0:
                    self.entropy -= frac * numpy.log2(frac)

            self.gain = None
            self.split = None
            self.fieldIndex = None
            self.field = None
            # for each field...
            for fieldIndex, field in enumerate(self.dataset.fields):
                # go to a function that finds the best choice for that field
                if field.tpe == numbers.Real:
                    gainTerm, split = self.numericalEntropyGainTerm(field)
                elif field.tpe == str:
                    gainTerm, split = self.categoricalEntropyGainTerm(field, self.maxSubsetSize)
                else:
                    raise RuntimeError

                # the gainTerm functions don't include this constant (entropy of the unsplit node)
                gainTerm += self.entropy

                # ... and keep track of the best field, best split
                if self.gain is None or gainTerm > self.gain:
                    self.gain = gainTerm
                    self.split = split
                    self.fieldIndex = fieldIndex
                    self.field = field

        else:
            raise RuntimeError

        # construct a new dataset by splitting the best field, best split
        if self.field.tpe == numbers.Real:
            passSelection = self.field.data <= self.split
        elif self.field.tpe == str:
            passSelection = numpy.in1d(self.field.data, self.split)
        failSelection = numpy.logical_not(passSelection)

        passDataset = Dataset([], self.dataset.names)
        for field in self.dataset.fields:
            passDataset.fields.append(field.select(passSelection))

        failDataset = Dataset([], self.dataset.names)
        for field in self.dataset.fields:
            failDataset.fields.append(field.select(failSelection))

        if len(passDataset) > 0 and len(failDataset) > 0:
            passPredictand = self.predictand.select(passSelection)
            failPredictand = self.predictand.select(failSelection)

            # create two new tree nodes, one with the data that pass the cut, the other with the data that fail
            self.passBranch = TreeNode(passDataset, passPredictand, self.maxSubsetSize)
            self.failBranch = TreeNode(failDataset, failPredictand, self.maxSubsetSize)

    def numericalEntropyGainTerm(self, field):
        """Split a numerical predictor in such a way that maximizes entropic gain above and below the threshold of the split."""

        # sort values so that we can use running sums (numpy.cumsum)
        sortedIndexes = numpy.argsort(field.data, kind="heapsort")

        # work with the predictor (values) and predictand (categories) for ascending values of the predictor
        values = field.data[sortedIndexes]
        categories = self.predictand.data[sortedIndexes]
        
        # for normalizing
        numInSelection = numpy.arange(1, self.datasetSize + 1, dtype=numpy.dtype(float))
        numNotInSelection = numpy.arange(self.datasetSize, 0, -1, dtype=numpy.dtype(float))

        # sum over the values array using Numpy iteration
        selectionEntropy = numpy.zeros(self.datasetSize, dtype=numpy.dtype(float))
        antiSelectionEntropy = numpy.zeros(self.datasetSize, dtype=numpy.dtype(float))

        # sum over categories using a Python for loop
        # (dataset pre-processing has ensured that the N categories are integers from 0 to N-1, so xrange)
        for category in range(len(self.predictand.intToStr)):
            # number of instances of this category for value <= cut
            numAtAndBeforeIndex = numpy.cumsum(categories == category)

            # probability of the category in the (value <= cut) data subset
            frac = numAtAndBeforeIndex / numInSelection

            # for entropy, it is convenient to define 0*log(0) as 0, but Numpy reports it as NaN
            selectionEntropyTerm = frac * numpy.log2(frac)
            selectionEntropyTerm[numpy.isnan(selectionEntropyTerm)] = 0.0

            selectionEntropy -= selectionEntropyTerm

            # now do the same for the anti-selection
            numAtAndBeforeIndex = numAtAndBeforeIndex[-1] - numAtAndBeforeIndex
            frac = numAtAndBeforeIndex / numNotInSelection
            selectionEntropyTerm = frac * numpy.log2(frac)
            selectionEntropyTerm[numpy.isnan(selectionEntropyTerm)] = 0.0

            antiSelectionEntropy -= selectionEntropyTerm

        # calculate gain for all possible index splits
        # (even if some are not unique--- it's faster to just have Numpy scan over them than to try to skip them)
        gains = -(numInSelection/self.datasetSize)*selectionEntropy - (numNotInSelection/self.datasetSize)*antiSelectionEntropy

        # find the highest gain, preferring larger indexes if there's a tie
        # (that's why we reverse the order of the list with [::-1] and then correct for the index for the reversal)
        maxGainIndex = len(gains) - 1 - numpy.argmax(gains[::-1])
        gainTerm = gains[maxGainIndex]

        # for niceness, put the cut between the best index (assuming the cut is "<=") and the next one,
        # so that the cut could equivalently be "<"
        # (the next higher should be exactly one higher because we picked the larger indexes in a tie above)
        nextHigher = maxGainIndex
        while nextHigher < self.datasetSize - 1 and values[nextHigher] == values[maxGainIndex]:
            nextHigher += 1
        cutValue = (values[maxGainIndex] + values[nextHigher]) / 2.0

        return gainTerm, cutValue

    def categoricalEntropyGainTerm(self, field, maxSubsetSize=None):
        """Split a categorical predictor in such a way that maximizes entropic gain inside and outside of a subset of predictor values.

        :type field: titus.producer.cart.Dataset.Field
        :param field: the field to consider when calculating the entropy gain term
        :type maxSubsetSize: positive integer or ``None``
        :param maxSubsetSize: maximum size of subset splits of categorical regressors (approximation for optimization in ``categoricalEntropyGainTerm`` and ``categoricalNVarianceGainTerm``)
        :rtype: (number, list of strings)
        :return: (best gain term, best combination of regressor categories)
        """

        # get a selection array for each category of the variable we want to use to make the prediciton
        numPredictorCategories = len(field.strToInt)
        predictorSelection = numpy.zeros((self.datasetSize, numPredictorCategories), dtype=numpy.dtype(bool))
        for predictorCategory in range(numPredictorCategories):
            predictorSelection[:,predictorCategory] = (field.data == predictorCategory)

        # get a selection array for each category that we want to predict
        numPredictandCategories = len(self.predictand.strToInt)
        predictandSelection = numpy.zeros((self.datasetSize, numPredictandCategories), dtype=numpy.dtype(bool))
        for predictandCategory in range(numPredictandCategories):
            predictandSelection[:,predictandCategory] = (self.predictand.data == predictandCategory)

        # combine them for all combinations of predictor category and predictand category
        # bitwise_and is equivalent to logical_and (for these boolean arrays) and faster
        numInCategory = [numpy.array([numpy.sum(numpy.bitwise_and(predictorSelection[:,i], predictandSelection[:,j]))
                          for i in range(numPredictorCategories)])
                         for j in range(numPredictandCategories)]

        # for the denominators
        numMarginal = numpy.array([numpy.sum(predictorSelection[:,i]) for i in range(numPredictorCategories)])

        # only consider categories that still have instances at this depth of the tree
        remainingPredictorCategories = [i for i in range(numPredictorCategories) if numMarginal[i] > 0]
        remainingPredictandCategories = numpy.array([i for i in range(numPredictandCategories) if self.predictandDistribution[i] > 0])

        # we will iterate over (N choose k) for all k from 1 to the highest informative k (half of N, rounding up)
        # or a user-supplied maxSubsetSize (to avoid very long calculations)
        maxInformativeSubset = int(math.ceil(len(remainingPredictorCategories) / 2.0))
        if maxSubsetSize is None:
            maxSubsetSize = maxInformativeSubset
        else:
            maxSubsetSize = min(maxSubsetSize, maxInformativeSubset)

        bestGainTerm = None
        bestCombination = None
        for howMany in range(1, maxSubsetSize + 1):
            # itertools.combinations does (N choose k) for a specific k
            for categorySet in itertools.combinations(remainingPredictorCategories, howMany):
                selection = numpy.array(categorySet)
                antiselection = numpy.setdiff1d(remainingPredictorCategories, selection, assume_unique=True)

                numInSelection = 1.0 * numpy.sum(numMarginal[selection])
                numInAntiselection = self.datasetSize - numInSelection

                selectionEntropy = 0.0
                antiSelectionEntropy = 0.0

                for predictandCategory in remainingPredictandCategories:
                    # for the selection
                    numInPredictand = numpy.sum(numInCategory[predictandCategory][selection])
                    if numInPredictand != 0:
                        frac = numInPredictand / numInSelection
                        selectionEntropy -= frac * numpy.log2(frac)

                    # now for the antiselection
                    numInPredictand = numpy.sum(numInCategory[predictandCategory][antiselection])
                    if numInPredictand != 0:
                        frac = numInPredictand / numInAntiselection
                        antiSelectionEntropy -= frac * numpy.log2(frac)

                # gainTerm is scalar, driven by a Python for loop over itertools.combinations
                gainTerm = -(numInSelection/self.datasetSize)*selectionEntropy - (numInAntiselection/self.datasetSize)*antiSelectionEntropy

                # find the best one
                if bestGainTerm is None or gainTerm > bestGainTerm:
                    bestGainTerm = gainTerm
                    bestCombination = categorySet

        return bestGainTerm, bestCombination

    def numericalNVarianceGainTerm(self, field):
        """Split a numerical predictor in such a way that maximizes n-times-variance gain above and below the threshold of the split.
        
        :type field: titus.producer.cart.Dataset.Field
        :param field: the field to consider when calculating the n-times variance gain term
        :rtype: (number, number)
        :return: (best gain term, best cut value)
        """

        # sort values so that we can use running sums (numpy.cumsum)
        sortedIndexes = numpy.argsort(field.data, kind="heapsort")

        # work with the predictor (values) and predictand (predictands) for ascending values of the predictor
        values = field.data[sortedIndexes]
        predictands = self.predictand.data[sortedIndexes]

        # compute less-than-or-equal-to sums for each index using Numpy
        sum1sel = numpy.arange(1, len(values) + 1, dtype=numpy.dtype(float))
        sumxsel = numpy.cumsum(predictands)
        sumxxsel = numpy.cumsum(numpy.power(predictands, 2))

        # derive greater-than sums from the total
        sum1antisel = sum1sel[-1] - sum1sel
        sumxantisel = sumxsel[-1] - sumxsel
        sumxxantisel = sumxxsel[-1] - sumxxsel

        # skip the last entry (run over arrays as [:-1]) because it's the only one whose denominator is zero
        nTimesVarianceInSelection = sumxxsel[:-1] - (sumxsel[:-1]**2/sum1sel[:-1])
        nTimesVarianceInAntiselection = sumxxantisel[:-1] - (sumxantisel[:-1]**2/sum1antisel[:-1])

        # calculate gain for all possible index splits
        # (even if some are not unique--- it's faster to just have Numpy scan over them than to try to skip them)
        gains = -nTimesVarianceInSelection - nTimesVarianceInAntiselection

        # find the highest gain, preferring larger indexes if there's a tie
        # (that's why we reverse the order of the list with [::-1] and then correct for the index for the reversal)
        maxGainIndex = len(gains) - 1 - numpy.argmax(gains[::-1])
        gainTerm = gains[maxGainIndex]

        # for niceness, put the cut between the best index (assuming the cut is "<=") and the next one,
        # so that the cut could equivalently be "<"
        # (the next higher should be exactly one higher because we picked the larger indexes in a tie above)
        nextHigher = maxGainIndex
        while nextHigher < self.datasetSize - 1 and values[nextHigher] == values[maxGainIndex]:
            nextHigher += 1
        cutValue = (values[maxGainIndex] + values[nextHigher]) / 2.0

        return gainTerm, cutValue

    def categoricalNVarianceGainTerm(self, field, maxSubsetSize=None):
        """Split a categorical predictor in such a way that maximizes n-times-variance gain inside and outside of a subset of predictor values.

        :type field: titus.producer.cart.Dataset.Field
        :param field: the field to consider when calculating the n-times-variance gain term
        :type maxSubsetSize: positive integer or ``None``
        :param maxSubsetSize: maximum size of subset splits of categorical regressors (approximation for optimization in ``categoricalEntropyGainTerm`` and ``categoricalNVarianceGainTerm``)
        :rtype: (number, list of strings)
        :return: (best gain term, best combination of regressor categories)
        """

        # find unique values of the predictor field
        try:
            fieldUniques = numpy.unique(field.data)
        except TypeError:
            fieldUniques = numpy.unique1d(field.data)

        # compute a partial sum for each unique value
        sum1 = numpy.empty(len(fieldUniques), dtype=numpy.dtype(float))
        sumx = numpy.empty(len(fieldUniques), dtype=numpy.dtype(float))
        sumxx = numpy.empty(len(fieldUniques), dtype=numpy.dtype(float))
        for i, v in enumerate(fieldUniques):
            predictandValues = self.predictand.data[field.data == v]
            sum1[i] = len(predictandValues)
            sumx[i] = numpy.sum(predictandValues)
            sumxx[i] = numpy.sum(numpy.power(predictandValues, 2))

        # get the totals to more quickly compute antiselections
        sum1total = numpy.sum(sum1)
        sumxtotal = numpy.sum(sumx)
        sumxxtotal = numpy.sum(sumxx)

        # we will iterate over (N choose k) for all k from 1 to the highest informative k (half of N, rounding up)
        # or a user-supplied maxSubsetSize (to avoid very long calculations)
        maxInformativeSubset = int(math.ceil(len(fieldUniques) / 2.0))
        if maxSubsetSize is None:
            maxSubsetSize = maxInformativeSubset
        else:
            maxSubsetSize = min(maxSubsetSize, maxInformativeSubset)

        bestGainTerm = None
        bestCombination = None
        for howMany in range(1, maxSubsetSize + 1):
            # itertools.combinations does (N choose k) for a specific k
            for categorySet in itertools.combinations(range(len(fieldUniques)), howMany):
                selection = numpy.array(categorySet)

                # for the selection
                sum1sel = numpy.sum(sum1[selection])
                sumxsel = numpy.sum(sumx[selection])
                sumxxsel = numpy.sum(sumxx[selection])
                nTimesVarianceInSelection = (sumxxsel - sumxsel**2/sum1sel) if sum1sel > 0.0 else 0.0

                # for the antiselection
                sum1antisel = sum1total - sum1sel
                sumxantisel = sumxtotal - sumxsel
                sumxxantisel = sumxxtotal - sumxxsel
                nTimesVarianceInAntiselection = (sumxxantisel - sumxantisel**2/sum1antisel) if sum1antisel > 0.0 else 0.0

                # gainTerm is scalar, driven by a Python for loop over itertools.combinations
                gainTerm = -nTimesVarianceInSelection - nTimesVarianceInAntiselection

                # find the best one
                if bestGainTerm is None or gainTerm > bestGainTerm:
                    bestGainTerm = gainTerm
                    bestCombination = categorySet

        return bestGainTerm, bestCombination

    def walkNodes(self, topDown=True, depth=0):
        """Return a generator that walks over all nodes in the tree, yielding a 2-tuple of node and depth.

        :type topDown: bool
        :param topDown: if ``True``, do a depth-first walk from root to leaf; if ``False``, do a depth-first walk from leaf to root
        :type depth: int
        :param depth: starting value for reporting depth
        :rtype: generator of (titus.producer.cart.TreeNode, int)
        :return: generator of (node, depth)
        """

        if topDown:
            yield self, depth
        if hasattr(self, "passBranch"):
            for n, d in self.passBranch.walkNodes(topDown, depth + 1):
                yield n, d
        if hasattr(self, "failBranch"):
            for n, d in self.failBranch.walkNodes(topDown, depth + 1):
                yield n, d
        if not topDown:
            yield self, depth

    def walkLeaves(self):
        """Return a generator that walks over all leaves in the tree, yielding a 2-tuple of node and depth.

        :rtype: generator of (titus.producer.cart.TreeNode, int)
        :return: generator of (node, depth)
        """

        for node, depth in self.walkNodes():
            if not hasattr(node, "passBranch") and not hasattr(node, "failBranch"):
                yield node, depth

    def pfaValueType(self, dataType):
        """Create an Avro schema representing the comparison value type.

        :type dataType: Pythonized JSON
        :param dataType: Avro record schema of the input data
        :rtype: Pythonized JSON
        :return: value type (``value`` field of the PFA ``TreeNode``)
        """

        if dataType.get("type", None) != "record":
            raise TypeError("dataType must be a record")

        dataTypeName = dataType["name"]
        dataFieldNames = [x["name"] for x in dataType["fields"]]

        if not set(self.dataset.names).issubset(set(dataFieldNames)):
            raise TypeError("dataType must be a record with at least as many fields as the ones used to train the tree")

        dataFieldTypes = []
        for field in dataType["fields"]:
            try:
                fieldIndex = self.dataset.names.index(field["name"])
            except ValueError:
                pass
            else:
                if self.dataset.fields[fieldIndex].tpe == numbers.Real:
                    if field["type"] not in ("int", "long", "float", "double"):
                        raise TypeError("dataType field \"{0}\" must be a numeric type, since this was a numeric type in the dataset training".format(field["name"]))
                    dataFieldTypes.append(field["type"])
                elif self.dataset.fields[fieldIndex].tpe == str:
                    if field["type"] != "string":
                        raise TypeError("dataType field \"{0}\" must be a string, since this was a string in the dataset training".format(field["name"]))
                    if self.maxSubsetSize == 1:
                        dataFieldTypes.append(field["type"])
                    else:
                        dataFieldTypes.append({"type": "array", "items": field["type"]})

        asjson = [json.dumps(x) for x in dataFieldTypes]
        astypes = ForwardDeclarationParser().parse(asjson)
        return LabelData.broadestType(list(astypes.values()))

    def pfaScoreType(self):
        """Create an Avro schema representing the score type.

        :rtype: Pythonized JSON
        :return: score type (part of the ``pass`` and ``fail`` unions of the PFA ``TreeNode``)
        """

        if self.predictand.tpe == numbers.Real:
            return "double"
        elif self.predictand.tpe == str:
            return "string"
        else:
            raise RuntimeError

    def pfaType(self, dataType, treeTypeName, nodeScores=False, datasetSize=False, predictandDistribution=False, predictandUnique=False, entropy=False, nTimesVariance=False, gain=False):
        """Create a PFA type schema representing this tree.

        :type dataType: Pythonized JSON
        :param dataType: Avro record schema of the input data
        :type treeTypeName: string
        :param treeTypeName: name of the tree node record (usually ``TreeNode``)
        :type nodeScores: bool
        :param nodeScores: if ``True``, include a field for intermediate node scores
        :type datasetSize: bool
        :param datasetSize: if ``True``, include a field for the size of the training dataset at each node
        :type predictandDistribution: bool
        :param predictandDistribution: if ``True``, include a field for the distribution of training predictand values (only for classification trees)
        :type predictandUnique: bool
        :param predictandUnique: if ``True``, include a field for unique predictand values at each node
        :type entropy: bool
        :param entropy: if ``True``, include an entropy term at each node (only for classification trees)
        :type nTimesVariance: bool
        :param nTimesVariance: if ``True``, include an n-times-variance term at each node (only for regression trees)
        :type gain: bool
        :param gain: if ``True``, include a gain term at each node
        :rtype: Pythonized JSON
        :return: Avro schema for the tree node type
        """

        valueType = self.pfaValueType(dataType)
        scoreType = self.pfaScoreType()

        out = OrderedDict([("type", "record"),
                           ("name", treeTypeName),
                           ("fields", [{"name": "field", "type": {"type": "enum",
                                                                  "name": dataType["name"] + "Fields",
                                                                  "symbols": [x["name"] for x in dataType["fields"]]}},
                                       {"name": "operator", "type": "string"},
                                       {"name": "value", "type": json.loads(str(valueType.schema))},
                                       {"name": "pass", "type": [scoreType, treeTypeName]},
                                       {"name": "fail", "type": [scoreType, treeTypeName]}])])
        
        if nodeScores:
            out["fields"].append({"name": "score", "type": scoreType})

        if datasetSize:
            out["fields"].append({"name": "datasetSize", "type": "int"})

        if predictandDistribution:
            if self.predictand.tpe != str:
                raise TypeError("predictandDistribution can only be used if the predictand is a string (classification trees)")
            out["fields"].append({"name": "scoreDistribution", "type": {"type": "map", "values": "int"}})

        if predictandUnique:
            out["fields"].append({"name": "scoreValues", "type": {"type": "array", "items": scoreType}})

        if entropy:
            if self.predictand.tpe != str:
                raise TypeError("entropy can only be used if the predictand is a string (classification trees)")
            out["fields"].append({"name": "entropy", "type": "double"})

        if nTimesVariance:
            if self.predictand.tpe != numbers.Real:
                raise TypeError("nTimesVariance can only be used if the predictand is a number (regression trees)")
            out["fields"].append({"name": "nTimesVariance", "type": "double"})

        if gain:
            out["fields"].append({"name": "gain", "type": "double"})

        return out

    def splitField(self):
        """Return the name of the input field at this split or ``None`` if this is a leaf node."""
        if hasattr(self, "fieldIndex"):
            return self.dataset.names[self.fieldIndex]
        else:
            return None

    def pfaValue(self, dataType, treeTypeName, nodeScores=False, datasetSize=False, predictandDistribution=False, predictandUnique=False, entropy=False, nTimesVariance=False, gain=False, valueType=None):
        """Create a PFA data structure representing this tree.

        :type dataType: Pythonized JSON
        :param dataType: Avro record schema of the input data
        :type treeTypeName: string
        :param treeTypeName: name of the tree node record (usually ``TreeNode``)
        :type nodeScores: bool
        :param nodeScores: if ``True``, include a field for intermediate node scores
        :type datasetSize: bool
        :param datasetSize: if ``True``, include a field for the size of the training dataset at each node
        :type predictandDistribution: bool
        :param predictandDistribution: if ``True``, include a field for the distribution of training predictand values (only for classification trees)
        :type predictandUnique: bool
        :param predictandUnique: if ``True``, include a field for unique predictand values at each node
        :type entropy: bool
        :param entropy: if ``True``, include an entropy term at each node (only for classification trees)
        :type nTimesVariance: bool
        :param nTimesVariance: if ``True``, include an n-times-variance term at each node (only for regression trees)
        :type gain: bool
        :param gain: if ``True``, include a gain term at each node
        :type valueType: Pythonized JSON or ``None``
        :param valueType: if ``None``, call ``self.pfaValueType(dataType)`` to generate a value type; otherwise, take the given value
        :rtype: Pythonized JSON
        :return: PFA data structure for the tree, to be inserted into the cell or pool's ``init`` field
        """

        scoreType = self.pfaScoreType()
            
        if not hasattr(self, "passBranch") and not hasattr(self, "failBranch"):
            return {scoreType: self.score()}

        else:
            fieldName = self.dataset.names[self.fieldIndex]

            if self.field.tpe == numbers.Real:
                operator = "<="
                value = self.split
            elif self.field.tpe == str:
                if self.maxSubsetSize == 1:
                    operator = "=="
                    value = self.dataset.fields[self.fieldIndex].intToStr[self.split[0]]
                else:
                    operator = "in"
                    value = [self.dataset.fields[self.fieldIndex].intToStr[x] for x in self.split]
            else:
                raise RuntimeError

            if valueType is None:
                valueType = self.pfaValueType(dataType)
            if isinstance(valueType, AvroUnion):
                specificType, = [x["type"] for x in dataType["fields"] if x["name"] == fieldName]
                if operator == "in" or operator == "notIn":
                    specificType = {"type": "array", "items": specificType}
                if isinstance(specificType, dict):
                    specificType = specificType["type"]
                value = {specificType: value}
            
            passBranch = self.passBranch.pfaValue(dataType, treeTypeName, nodeScores, datasetSize, predictandDistribution, predictandUnique, entropy, nTimesVariance, gain, valueType)
            failBranch = self.failBranch.pfaValue(dataType, treeTypeName, nodeScores, datasetSize, predictandDistribution, predictandUnique, entropy, nTimesVariance, gain, valueType)

            if list(passBranch.keys()) != [scoreType]:
                passBranch = {treeTypeName: passBranch}
            if list(failBranch.keys()) != [scoreType]:
                failBranch = {treeTypeName: failBranch}

            out = OrderedDict([("field", fieldName),
                               ("operator", operator),
                               ("value", value),
                               ("pass", passBranch),
                               ("fail", failBranch)])

            if nodeScores:
                out["score"] = self.score()

            if datasetSize:
                out["datasetSize"] = self.datasetSize

            if predictandDistribution:
                if self.predictand.tpe != str:
                    raise TypeError("predictandDistribution can only be used if the predictand is a string (classification trees)")
                out["scoreDistribution"] = OrderedDict(sorted((self.predictand.intToStr[i], x) for i, x in enumerate(self.predictandDistribution)))

            if predictandUnique:
                if hasattr(self, "predictandDistribution"):
                    out["scoreValues"] = sorted(self.predictand.intToStr[i] for i, x in enumerate(self.predictandDistribution) if x > 0.0)
                elif hasattr(self, "predictandUnique"):
                    out["scoreValues"] = sorted(self.predictandUnique)
                else:
                    raise RuntimeError

            if entropy:
                if self.predictand.tpe != str:
                    raise TypeError("entropy can only be used if the predictand is a string (classification trees)")
                out["entropy"] = self.entropy

            if nTimesVariance:
                if self.predictand.tpe != numbers.Real:
                    raise TypeError("nTimesVariance can only be used if the predictand is a number (regression trees)")
                out["nTimesVariance"] = self.nTimesVariance

            if gain:
                out["gain"] = self.gain

            return out

    def pfaDocument(self, inputType, treeTypeName, dataType=None, preprocess=None, nodeScores=False, datasetSize=False, predictandDistribution=False, predictandUnique=False, entropy=False, nTimesVariance=False, gain=False):
        """Create a PFA document to score with this tree.

        :type inputType: Pythonized JSON
        :param inputType: Avro record schema of the input data
        :type treeTypeName: string
        :param treeTypeName: name of the tree node record (usually ``TreeNode``)
        :type dataType: Pythonized JSON
        :param dataType: Avro record schema of the data that goes to the tree, possibly preprocessed
        :type preprocess: PrettyPFA substitution or ``None``
        :param preprocess: pre-processing expression
        :type nodeScores: bool
        :param nodeScores: if ``True``, include a field for intermediate node scores
        :type datasetSize: bool
        :param datasetSize: if ``True``, include a field for the size of the training dataset at each node
        :type predictandDistribution: bool
        :param predictandDistribution: if ``True``, include a field for the distribution of training predictand values (only for classification trees)
        :type predictandUnique: bool
        :param predictandUnique: if ``True``, include a field for unique predictand values at each node
        :type entropy: bool
        :param entropy: if ``True``, include an entropy term at each node (only for classification trees)
        :type nTimesVariance: bool
        :param nTimesVariance: if ``True``, include an n-times-variance term at each node (only for regression trees)
        :type gain: bool
        :param gain: if ``True``, include a gain term at each node
        :rtype: Pythonized JSON
        :return: complete PFA document for running tree classification or regression
        """

        if dataType is None:
            dataType = inputType
        dataTypeName = dataType["name"]

        scoreType = self.pfaScoreType()
        treeType = self.pfaType(dataType, treeTypeName, nodeScores, datasetSize, predictandDistribution, predictandUnique, entropy, nTimesVariance, gain)
        treeValue = self.pfaValue(dataType, treeTypeName, nodeScores, datasetSize, predictandDistribution, predictandUnique, entropy, nTimesVariance, gain)
        if preprocess is None:
            preprocess = "input"

        return titus.prettypfa.jsonNode('''
input: <<inputType>>
output: <<scoreType>>
cells:
  tree(<<treeType>>) = <<treeValue>>
action:
  model.tree.simpleWalk(<<preprocess>>, tree,
    fcn(d: <<dataTypeName>>, t: <<treeTypeName>> -> boolean)
      model.tree.simpleTest(d, t))
''', **vars())



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import *
from titus.util import callfcn, div
import titus.P as P
from titus.lib.array import argLowestN

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "model.neighbor."

#################################################################### 

class Mean(LibFcn):
    name = prefix + "mean"
    sig = Sigs([Sig([{"points": P.Array(P.Array(P.Double()))}], P.Array(P.Double())),
                Sig([{"points": P.Array(P.Array(P.Double()))}, {"weight": P.Fcn([P.Array(P.Double())], P.Double())}], P.Array(P.Double()))])
    errcodeBase = 30000
    def __call__(self, state, scope, pos, paramTypes, points, *args):
        if len(args) == 1:
            weight, = args
        else:
            weight = None

        if len(points) == 0:
            raise PFARuntimeException("not enough points", self.errcodeBase + 0, self.name, pos)

        dimensions = len(points[0])
        numer = [0.0] * dimensions
        denom = [0.0] * dimensions
        for point in points:
            if len(point) != dimensions:
                raise PFARuntimeException("inconsistent dimensionality", self.errcodeBase + 1, self.name, pos)

            if weight is None:
                w = 1.0
            else:
                w = callfcn(state, scope, weight, [point])

            for i in range(dimensions):
                numer[i] += w * point[i]
                denom[i] += w

        for i in range(dimensions):
            numer[i] /= denom[i]
        return numer
provide(Mean())

class NearestK(LibFcn):
    name = prefix + "nearestK"
    sig = Sigs([
        Sig([{"k": P.Int()}, {"datum": P.Array(P.Double())}, {"codebook": P.Array(P.Array(P.Double()))}], P.Array(P.Array(P.Double()))),
        Sig([{"k": P.Int()}, {"datum": P.Wildcard("A")}, {"codebook": P.Array(P.Wildcard("B"))}, {"metric": P.Fcn([P.Wildcard("A"), P.Wildcard("B")], P.Double())}], P.Array(P.Wildcard("B")))])

    errcodeBase = 30010
    def __call__(self, state, scope, pos, paramTypes, k, datum, codebook, *args):
        if k < 0:
            raise PFARuntimeException("k must be nonnegative", self.errcodeBase + 0, self.name, pos)

        if len(args) == 1:
            metric, = args
            distances = [callfcn(state, scope, metric, [datum, x]) for x in codebook]
        else:
            if len(codebook) == 0:
                return []
            else:
                dimensions = len(datum)
                for x in codebook:
                    if len(x) != dimensions:
                        raise PFARuntimeException("inconsistent dimensionality", self.errcodeBase + 1, self.name, pos)
            distances = [sum((di - xi)**2 for di, xi in zip(datum, x)) for x in codebook]

        indexes = argLowestN(distances, k, lambda a, b: a < b)
        return [codebook[i] for i in indexes]

provide(NearestK())

class BallR(LibFcn):
    name = prefix + "ballR"
    sig = Sigs([
        Sig([{"r": P.Double()}, {"datum": P.Array(P.Double())}, {"codebook": P.Array(P.Array(P.Double()))}], P.Array(P.Array(P.Double()))),
        Sig([{"r": P.Double()}, {"datum": P.Wildcard("A")}, {"codebook": P.Array(P.Wildcard("B"))}, {"metric": P.Fcn([P.Wildcard("A"), P.Wildcard("B")], P.Double())}], P.Array(P.Wildcard("B")))])
    errcodeBase = 30020
    def __call__(self, state, scope, pos, paramTypes, r, datum, codebook, *args):
        if len(args) == 1:
            metric, = args
            distances = [callfcn(state, scope, metric, [datum, x]) for x in codebook]
        else:
            distances = [math.sqrt(sum((di - xi)**2 for di, xi in zip(datum, x))) for x in codebook]
        return [x for x, d in zip(codebook, distances) if d < r]
provide(BallR())



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import io

import titus.producer.tools as t
from titus.inspector.defs import *

def depthGreaterThan(obj, target):
    """Helper function for determining if an object's depth is greater than a given target.

    A string, number, boolean, or null has depth 0, a list or dict of such objects has depth 1, etc.

    :type obj: Pythonized JSON
    :param obj: object to inspect
    :type target: non-negative integer
    :param target: target depth
    :rtype: bool
    :return: ``True`` if the depth of ``obj`` is greater than ``target``; ``False`` otherwise.
    """

    if isinstance(obj, dict):
        return any(depthGreaterThan(x, target - 1) for x in list(obj.values())) or \
               (len(obj) == 0 and target <= 0)
    elif isinstance(obj, (list, tuple)):
        return any(depthGreaterThan(x, target - 1) for x in obj) or \
               (len(obj) == 0 and target <= 0)
    else:
        return target < 0
        
class LookCommand(Command):
    """The 'json look' command in pfainspector."""

    def __init__(self, mode):
        self.name = "look"
        self.syntax = "look <name> [maxDepth=8] [indexWidth=30]"
        self.help = "look at a named PFA document or subexpression in memory\n    " + self.syntax
        self.mode = mode

    def complete(self, established, active):
        """Handle tab-complete for this command's arguments.

        :type established: string
        :param established: part of the text that has been established
        :type active: string
        :param active: part of the text to be completed
        :rtype: list of strings
        :return: potential completions
        """

        options = ["maxDepth=", "indexWidth="]
        words = getcomplete(established)

        if len(words) == 0:
            if active in self.mode.pfaFiles:
                return [active + "["]
            else:
                return sorted(x for x in self.mode.pfaFiles if x.startswith(active))

        elif len(words) == 1 and isinstance(words[0], parser.Extract) and words[0].partial:
            if words[0].text in self.mode.pfaFiles:
                return [x for x in extcomplete(self.mode.pfaFiles[words[0].text].obj, words[0].items) if x.startswith(active)]
            else:
                return []

        elif not words[-1].partial:
            return [x for x in options if x.startswith(active)]

        else:
            return []

    def action(self, args):
        """Perform the action associated with this command.

        :type args: list of titus.inspector.parser.Ast
        :param args: arguments passed to the command
        :rtype: ``None``
        :return: nothing; results must be printed to the screen
        """

        if len(args) == 1 and args[0] == parser.Word("help"):
            print(self.help)
        else:
            options = {"maxDepth": 8, "indexWidth": 30}
            while len(args) > 0 and isinstance(args[-1], parser.Option):
                opt = args.pop()
                if opt.word.text in ["maxDepth", "indexWidth"]:
                    try:
                        options[opt.word.text] = opt.value.value()
                    except TypeError:
                        raise InspectorError("illegal value for {0}".format(opt.word.text))
                else:
                    raise InspectorError("option {0} unrecognized".format(opt.word.text))

            if not isinstance(options["maxDepth"], int) or options["maxDepth"] <= 0:
                raise InspectorError("maxDepth must be a positive integer")

            if not isinstance(options["indexWidth"], int) or options["indexWidth"] <= 0:
                raise InspectorError("indexWidth must be a positive integer")

            if len(args) == 1 and isinstance(args[0], parser.Word):
                if args[0].text not in self.mode.pfaFiles:
                    raise InspectorError("no PFA document named \"{0}\" in memory (try 'load <file> as {1}')".format(args[0].text, args[0].text))
                node = self.mode.pfaFiles[args[0].text].obj

            elif len(args) == 1 and isinstance(args[0], parser.Extract):
                if args[0].text not in self.mode.pfaFiles:
                    raise InspectorError("no PFA document named \"{0}\" in memory (try 'load <file> as {1}')".format(args[0].text, args[0].text))
                node = self.mode.pfaFiles[args[0].text].obj
                items = args[0].items
                node = extaction(args[0], node, items)

            else:
                self.syntaxError()

            if not depthGreaterThan(node, 0):
                print(json.dumps(node))

            else:
                content = io.StringIO()
                if not depthGreaterThan(node, 1):
                    t.look(node, maxDepth=options["maxDepth"], indexWidth=options["indexWidth"], inlineDepth=0, stream=content)
                elif not depthGreaterThan(node, 2):
                    t.look(node, maxDepth=options["maxDepth"], indexWidth=options["indexWidth"], inlineDepth=1, stream=content)
                else:
                    t.look(node, maxDepth=options["maxDepth"], indexWidth=options["indexWidth"], inlineDepth=2, stream=content)

                content = content.getvalue()
                if content.count("\n") <= 100:
                    print(content)
                else:
                    proc = pipe("less")
                    try:
                        proc.stdin.write(content)
                    except IOError as err:
                        if str(err) != "[Errno 32] Broken pipe":
                            raise
                    pipewait(proc)

class CountCommand(Command):
    """The 'json count' command in pfainspector."""

    def __init__(self, mode):
        self.name = "count"
        self.syntax = "count <name> <pattern>"
        self.help = "count instances in a PFA document or subexpression that match a regular expression\n    " + self.syntax
        self.mode = mode

    def complete(self, established, active):
        """Handle tab-complete for this command's arguments.

        :type established: string
        :param established: part of the text that has been established
        :type active: string
        :param active: part of the text to be completed
        :rtype: list of strings
        :return: potential completions
        """

        words = getcomplete(established)

        if len(words) == 0:
            if active in self.mode.pfaFiles:
                return [active + "["]
            else:
                return sorted(x for x in list(self.mode.pfaFiles.keys()) if x.startswith(active))

        elif len(words) == 1 and isinstance(words[0], parser.Extract) and words[0].partial:
            if words[0].text in self.mode.pfaFiles:
                return [x for x in extcomplete(self.mode.pfaFiles[words[0].text].obj, words[0].items) if x.startswith(active)]
            else:
                return []

        else:
            return []

    def action(self, args):
        """Perform the action associated with this command.

        :type args: list of titus.inspector.parser.Ast
        :param args: arguments passed to the command
        :rtype: ``None``
        :return: nothing; results must be printed to the screen
        """

        if len(args) == 1 and args[0] == parser.Word("help"):
            print(self.help)
        else:
            if len(args) == 2 and isinstance(args[0], parser.Word):
                if args[0].text not in self.mode.pfaFiles:
                    raise InspectorError("no PFA document named \"{0}\" in memory (try 'load <file> as {1}')".format(args[0].text, args[0].text))
                node = self.mode.pfaFiles[args[0].text].obj

            elif len(args) == 2 and isinstance(args[0], parser.Extract):
                if args[0].text not in self.mode.pfaFiles:
                    raise InspectorError("no PFA document named \"{0}\" in memory (try 'load <file> as {1}')".format(args[0].text, args[0].text))
                node = self.mode.pfaFiles[args[0].text].obj
                items = args[0].items
                node = extaction(args[0], node, items)

            else:
                self.syntaxError()

            regex = args[-1].regex()
            print("{0} matches".format(t.count(regex, node)))

class IndexCommand(Command):
    """The 'json index' command in pfainspector."""

    def __init__(self, mode):
        self.name = "index"
        self.syntax = "index <name> <pattern>"
        self.help = "list indexes of a PFA document or subexpression that match a regular expression\n    " + self.syntax
        self.mode = mode

    def complete(self, established, active):
        """Handle tab-complete for this command's arguments.

        :type established: string
        :param established: part of the text that has been established
        :type active: string
        :param active: part of the text to be completed
        :rtype: list of strings
        :return: potential completions
        """

        words = getcomplete(established)

        if len(words) == 0:
            if active in self.mode.pfaFiles:
                return [active + "["]
            else:
                return sorted(x for x in list(self.mode.pfaFiles.keys()) if x.startswith(active))

        elif len(words) == 1 and isinstance(words[0], parser.Extract) and words[0].partial:
            if words[0].text in self.mode.pfaFiles:
                return [x for x in extcomplete(self.mode.pfaFiles[words[0].text].obj, words[0].items) if x.startswith(active)]
            else:
                return []

        else:
            return []

    def action(self, args):
        """Perform the action associated with this command.

        :type args: list of titus.inspector.parser.Ast
        :param args: arguments passed to the command
        :rtype: ``None``
        :return: nothing; results must be printed to the screen
        """

        if len(args) == 1 and args[0] == parser.Word("help"):
            print(self.help)
        else:
            if len(args) == 2 and isinstance(args[0], parser.Word):
                if args[0].text not in self.mode.pfaFiles:
                    raise InspectorError("no PFA document named \"{0}\" in memory (try 'load <file> as {1}')".format(args[0].text, args[0].text))
                node = self.mode.pfaFiles[args[0].text].obj

            elif len(args) == 2 and isinstance(args[0], parser.Extract):
                if args[0].text not in self.mode.pfaFiles:
                    raise InspectorError("no PFA document named \"{0}\" in memory (try 'load <file> as {1}')".format(args[0].text, args[0].text))
                node = self.mode.pfaFiles[args[0].text].obj
                items = args[0].items
                node = extaction(args[0], node, items)

            else:
                self.syntaxError()

            regex = args[-1].regex()
            def display(i):
                if isinstance(i, str):
                    if " " in i:
                        return json.dumps(i)
                    else:
                        return i
                else:
                    return str(i)

            print("Indexes that match the pattern:")
            count = 0
            for index in t.indexes(regex, node):
                print("    [" + ", ".join(display(i) for i in index) + "]")
                count += 1
            if count == 0:
                print("    (none)")

class FindCommand(Command):
    """The 'json find' command in pfainspector."""

    def __init__(self, mode):
        self.name = "find"
        self.syntax = "find <name> <pattern> [maxDepth=3] [indexWidth=30]"
        self.help = "show all matches of a regular expression in a PFA document or subexpression\n    " + self.syntax
        self.mode = mode

    def complete(self, established, active):
        """Handle tab-complete for this command's arguments.

        :type established: string
        :param established: part of the text that has been established
        :type active: string
        :param active: part of the text to be completed
        :rtype: list of strings
        :return: potential completions
        """

        options = ["maxDepth=", "indexWidth="]
        words = getcomplete(established)

        if len(words) == 0:
            if active in self.mode.pfaFiles:
                return [active + "["]
            else:
                return sorted(x for x in self.mode.pfaFiles if x.startswith(active))

        elif len(words) == 1 and isinstance(words[0], parser.Extract) and words[0].partial:
            if words[0].text in self.mode.pfaFiles:
                return [x for x in extcomplete(self.mode.pfaFiles[words[0].text].obj, words[0].items) if x.startswith(active)]
            else:
                return []

        elif not words[-1].partial:
            return [x for x in options if x.startswith(active)]

        else:
            return []

    def action(self, args):
        """Perform the action associated with this command.

        :type args: list of titus.inspector.parser.Ast
        :param args: arguments passed to the command
        :rtype: ``None``
        :return: nothing; results must be printed to the screen
        """

        if len(args) == 1 and args[0] == parser.Word("help"):
            print(self.help)
        else:
            options = {"maxDepth": 3, "indexWidth": 30}
            while len(args) > 0 and isinstance(args[-1], parser.Option):
                opt = args.pop()
                if opt.word.text in ["maxDepth", "indexWidth"]:
                    try:
                        options[opt.word.text] = opt.value.value()
                    except TypeError:
                        raise InspectorError("illegal value for {0}".format(opt.word.text))
                else:
                    raise InspectorError("option {1} unrecognized".format(opt.word.text))

            if not isinstance(options["maxDepth"], int) or options["maxDepth"] <= 0:
                raise InspectorError("maxDepth must be a positive integer")

            if not isinstance(options["indexWidth"], int) or options["indexWidth"] <= 0:
                raise InspectorError("indexWidth must be a positive integer")

            if len(args) == 2 and isinstance(args[0], parser.Word):
                if args[0].text not in self.mode.pfaFiles:
                    raise InspectorError("no PFA document named \"{0}\" in memory (try 'load <file> as {1}')".format(args[0].text, args[0].text))
                node = self.mode.pfaFiles[args[0].text].obj

            elif len(args) == 2 and isinstance(args[0], parser.Extract):
                if args[0].text not in self.mode.pfaFiles:
                    raise InspectorError("no PFA document named \"{0}\" in memory (try 'load <file> as {1}')".format(args[0].text, args[0].text))
                node = self.mode.pfaFiles[args[0].text].obj
                items = args[0].items
                node = extaction(args[0], node, items)

            else:
                self.syntaxError()

            regex = args[-1].regex()
            def display(i):
                if isinstance(i, str):
                    if " " in i:
                        return json.dumps(i)
                    else:
                        return i
                else:
                    return str(i)

            content = io.StringIO()
            count = 0
            for index in t.indexes(regex, node):
                content.write("At index [" + ", ".join(display(i) for i in index) + "]:\n")

                matched = t.get(node, index)

                if not depthGreaterThan(matched, 0):
                    content.write(json.dumps(matched) + "\n")
                elif not depthGreaterThan(matched, 1):
                    t.look(matched, maxDepth=options["maxDepth"], indexWidth=options["indexWidth"], inlineDepth=0, stream=content)
                elif not depthGreaterThan(matched, 2):
                    t.look(matched, maxDepth=options["maxDepth"], indexWidth=options["indexWidth"], inlineDepth=1, stream=content)
                else:
                    t.look(matched, maxDepth=options["maxDepth"], indexWidth=options["indexWidth"], inlineDepth=2, stream=content)

                content.write("\n")
                count += 1
            if count == 0:
                print("    (none)")

            content = content.getvalue()
            if content.count("\n") <= 100:
                print(content)
            else:
                proc = pipe("less")
                try:
                    proc.stdin.write(content)
                except IOError as err:
                    if str(err) != "[Errno 32] Broken pipe":
                        raise
                pipewait(proc)

class ChangeCommand(Command):
    """The 'json change' command in pfainspector."""

    def __init__(self, mode):
        self.name = "change"
        self.syntax = "change <name> <pattern> to <replacement>"
        self.help = "replace instances in a PFA document or subexpression that match a regular expression\n    " + self.syntax
        self.mode = mode

    def complete(self, established, active):
        """Handle tab-complete for this command's arguments.

        :type established: string
        :param established: part of the text that has been established
        :type active: string
        :param active: part of the text to be completed
        :rtype: list of strings
        :return: potential completions
        """

        words = getcomplete(established)

        if len(words) == 0:
            if active in self.mode.pfaFiles:
                return [active + "["]
            else:
                return sorted(x for x in list(self.mode.pfaFiles.keys()) if x.startswith(active))

        elif len(words) == 1 and isinstance(words[0], parser.Extract) and words[0].partial:
            if words[0].text in self.mode.pfaFiles:
                return [x for x in extcomplete(self.mode.pfaFiles[words[0].text].obj, words[0].items) if x.startswith(active)]
            else:
                return []

        elif len(words) == 2 and isinstance(words[0], (parser.Word, parser.Extract)) and words[0].text in self.mode.pfaFiles:
            return ["to "]

        else:
            return []

    def action(self, args):
        """Perform the action associated with this command.

        :type args: list of titus.inspector.parser.Ast
        :param args: arguments passed to the command
        :rtype: ``None``
        :return: nothing; results must be printed to the screen
        """

        if len(args) == 1 and args[0] == parser.Word("help"):
            print(self.help)
        else:
            if len(args) == 4 and isinstance(args[0], parser.Word):
                if args[0].text not in self.mode.pfaFiles:
                    raise InspectorError("no PFA document named \"{0}\" in memory (try 'load <file> as {1}')".format(args[0].text, args[0].text))
                model = self.mode.pfaFiles[args[0].text]
                node = model.obj
                safecopy = copy.deepcopy(node)
                def rollback():
                    self.mode.pfaFiles[args[0].text].obj = safecopy
                regex = args[1].regex()
                replacement = args[3].replacement()

            elif len(args) == 4 and isinstance(args[0], parser.Extract):
                if args[0].text not in self.mode.pfaFiles:
                    raise InspectorError("no PFA document named \"{0}\" in memory (try 'load <file> as {1}')".format(args[0].text, args[0].text))
                model = self.mode.pfaFiles[args[0].text]
                node = model.obj
                safecopy = copy.deepcopy(node)
                def rollback():
                    self.mode.pfaFiles[args[0].text].obj = safecopy
                items = args[0].items
                node = extaction(args[0], node, items)
                regex = args[1].regex()
                replacement = args[3].replacement()

            else:
                self.syntaxError()

            def display(i):
                if isinstance(i, str):
                    if " " in i:
                        return json.dumps(i)
                    else:
                        return i
                else:
                    return str(i)

            def replace(value, groups):
                if isinstance(value, parser.Replacement):
                    try:
                        return groups[value.name]
                    except KeyError:
                        raise InspectorError("group ({0}) not found in regular expression".format(value.name))

                elif isinstance(value, dict):
                    return dict((k, replace(v, groups)) for k, v in list(value.items()))

                elif isinstance(value, (list, tuple)):
                    return [replace(x, groups) for x in value]

                else:
                    return value

            def removeAts(obj):
                if isinstance(obj, dict):
                    return dict((k, removeAts(v)) for k, v in list(obj.items()) if k != "@")
                elif isinstance(obj, (list, tuple)):
                    return [removeAts(x) for x in obj]
                else:
                    return obj

            ask = True
            self.mode.pause()
            try:
                for index, match in t.search(regex, node):
                    replacedReplacement = replace(replacement, match.groups)

                    if ask:
                        print("At index [" + ", ".join(display(i) for i in index) + "]:")
                        print("Original:  " + json.dumps(removeAts(t.get(node, index))))
                        print("Change to: " + json.dumps(replacedReplacement))
                        action = None
                        while action is None:
                            response = input("(Y/n/all/stop/revert): ")
                            normalized = response.strip().lower()
                            if normalized in ("", "y", "yes"):
                                action = "yes"
                            elif normalized in ("n", "no"):
                                action = "no"
                            elif normalized == "all":
                                action = "all"
                            elif normalized == "stop":
                                action = "stop"
                            elif normalized == "revert":
                                action = "revert"
                        print()

                    else:
                        action = "yes"

                    if action == "yes":
                        t.assign(node, index, replacedReplacement)
                    elif action == "all":
                        t.assign(node, index, replacedReplacement)
                        ask = False
                    elif action == "stop":
                        break
                    elif action == "revert":
                        rollback()
                        break

            except:
                rollback()
                self.mode.resume()
                raise

            else:
                model.reset()
                self.mode.resume()

class JsonGadget(Gadget):
    """The 'json' gadget in pfainspector."""

    def __init__(self, mode):
        self.commandGroup = CommandGroup("json", [
            LookCommand(mode),
            CountCommand(mode),
            IndexCommand(mode),
            FindCommand(mode),
            ChangeCommand(mode)
            ])



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest

from titus.pfaast import *
from titus.reader import jsonToAst
from titus.datatype import *
from titus.util import bytesToString

class TestJsonToAst(unittest.TestCase):
    def testEngineConfig(self):
        self.assertEqual(
            EngineConfig(
            "test",
            Method.MAP,
            AvroInt(),
            AvroString(),
            [],
            [Call("+", [LiteralInt(2), LiteralInt(2)])],
            [],
            {},
            None,
            None,
            {},
            {},
            None,
            None,
            None,
            {},
            {}),
            jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": {"+": [2, 2]}
}'''))

        self.assertEqual(
            EngineConfig(
            "test",
            Method.EMIT,
            AvroInt(),
            AvroString(),
            [],
            [Call("+", [LiteralInt(2), LiteralInt(2)])],
            [],
            {},
            None,
            None,
            {},
            {},
            None,
            None,
            None,
            {},
            {}),
            jsonToAst('''{
  "name": "test",
  "method": "emit",
  "input": "int",
  "output": "string",
  "action": [{"+": [2, 2]}]
}'''))

        self.assertEqual(
            EngineConfig(
            "test",
            Method.MAP,
            AvroInt(),
            AvroString(),
            [Call("+", [LiteralInt(2), LiteralInt(2)])],
            [Call("+", [LiteralInt(2), LiteralInt(2)])],
            [Call("+", [LiteralInt(2), LiteralInt(2)])],
            {},
            None,
            None,
            {},
            {},
            None,
            None,
            None,
            {},
            {}),
            jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "begin": [{"+": [2, 2]}],
  "action": [{"+": [2, 2]}],
  "end": [{"+": [2, 2]}]
}'''))

        self.assertEqual(
            EngineConfig(
            "test",
            Method.MAP,
            AvroInt(),
            AvroString(),
            [],
            [Call("+", [LiteralInt(2), LiteralInt(2)])],
            [],
            {"f": FcnDef([{"x": AvroInt()}, {"y": AvroString()}], AvroNull(), [LiteralNull()])},
            None,
            None,
            {},
            {},
            None,
            None,
            None,
            {},
            {}),
            jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"+": [2, 2]}],
  "fcns": {"f": {"params":[{"x":"int"},{"y":"string"}],"ret":"null","do":[null]}}
}'''))

        self.assertEqual(
            EngineConfig(
            "test",
            Method.MAP,
            AvroInt(),
            AvroString(),
            [],
            [Call("+", [LiteralInt(2), LiteralInt(2)])],
            [],
            {},
            None,
            None,
            {"private": Cell(AvroInt(), "0", False, False, CellPoolSource.EMBEDDED)},
            {"private": Pool(AvroInt(), {}, False, False, CellPoolSource.EMBEDDED)},
            None,
            None,
            None,
            {},
            {}),
            jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"+": [2, 2]}],
  "cells":{"private":{"type":"int","init":0,"shared":false,"rollback":false}},
  "pools":{"private":{"type":"int","init":{},"shared":false,"rollback":false}}
}'''))

        self.assertEqual(
            EngineConfig(
            "test",
            Method.MAP,
            AvroInt(),
            AvroString(),
            [],
            [Call("+", [LiteralInt(2), LiteralInt(2)])],
            [],
            {},
            None,
            None,
            {},
            {},
            12345,
            "hello",
            None,
            {"internal": "data"},
            {"param": json.loads("3")}),
            jsonToAst('''{
  "name": "test",
  "method": "map",
  "input": "int",
  "output": "string",
  "action": [{"+": [2, 2]}],
  "randseed":12345,
  "doc":"hello",
  "metadata":{"internal":"data"},
  "options":{"param":3}
}'''))

        self.assertEqual(
            EngineConfig(
            "test",
            Method.MAP,
            AvroInt(),
            AvroString(),
            [Call("+", [LiteralInt(2), LiteralInt(2)])],
            [Call("+", [LiteralInt(2), LiteralInt(2)])],
            [Call("+", [LiteralInt(2), LiteralInt(2)])],
            {"f": FcnDef([{"x": AvroInt()}, {"y": AvroString()}], AvroNull(), [LiteralNull()])},
            None,
            None,
            {"private": Cell(AvroInt(), "0", False, False, CellPoolSource.EMBEDDED)},
            {"private": Pool(AvroInt(), {}, False, False, CellPoolSource.EMBEDDED)},
            12345,
            "hello",
            None,
            {"internal": "data"},
            {"param": json.loads("3")}),
            jsonToAst('''{
  "name": "test",
  "method": "map",
  "input": "int",
  "output": "string",
  "begin": [{"+": [2, 2]}],
  "action": [{"+": [2, 2]}],
  "end": [{"+": [2, 2]}],
  "fcns": {"f": {"params":[{"x":"int"},{"y":"string"}],"ret":"null","do":[null]}},
  "cells":{"private":{"type":"int","init":0,"shared":false,"rollback":false}},
  "pools":{"private":{"type":"int","init":{},"shared":false,"rollback":false}},
  "randseed":12345,
  "doc":"hello",
  "metadata":{"internal":"data"},
  "options":{"param":3}
}'''))

    def testCell(self):
        self.assertEqual(
            EngineConfig(
            "test",
            Method.MAP,
            AvroInt(),
            AvroString(),
            [],
            [Call("+", [LiteralInt(2), LiteralInt(2)])],
            [],
            {},
            None,
            None,
            {"private": Cell(AvroArray(AvroString()), "[]", False, False, CellPoolSource.EMBEDDED)},
            {},
            None,
            None,
            None,
            {},
            {}),
            jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"+": [2, 2]}],
  "cells":{"private":{"type":{"type": "array", "items": "string"},"init":[],"shared":false,"rollback":false}}
}'''))

    def testDefineFunction(self):
        self.assertEqual(
            EngineConfig(
            "test",
            Method.MAP,
            AvroInt(),
            AvroString(),
            [],
            [Call("+", [LiteralInt(2), LiteralInt(2)])],
            [],
            {"f": FcnDef([{"x": AvroInt()}, {"y": AvroString()}], AvroNull(), [LiteralNull()])},
            None,
            None,
            {},
            {},
            None,
            None,
            None,
            {},
            {}),
            jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"+": [2, 2]}],
  "fcns": {"f": {"params":[{"x":"int"},{"y":"string"}],"ret":"null","do":[null]}}
}'''))

    def testCallFunction(self):
        self.assertEqual(
            EngineConfig(
            "test",
            Method.MAP,
            AvroInt(),
            AvroString(),
            [],
            [Call("+", [LiteralInt(2), LiteralInt(2)])],
            [],
            {},
            None,
            None,
            {},
            {},
            None,
            None,
            None,
            {},
            {}),
            jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"+": [2, 2]}]
}'''))

    def testCallWithFcn(self):
        self.assertEqual(
            EngineConfig(
            "test",
            Method.MAP,
            AvroInt(),
            AvroString(),
            [],
            [Call("sort", [Ref("array"), FcnRef("byname")])],
            [],
            {},
            None,
            None,
            {},
            {},
            None,
            None,
            None,
            {},
            {}),
            jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"sort":["array",{"fcn": "byname"}]}]
}'''))

    def testCallWithFcnDef(self):
        x = EngineConfig("test", Method.MAP, AvroInt(), AvroString(), [], [Call("sort", [Ref("array"), FcnDef([{"x": AvroInt()}, {"y": AvroString()}], AvroNull(), [LiteralNull()])])], [], {}, None, None, {}, {}, None, None, None, {}, {})
        y = jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"sort":["array",{"params": [{"x": "int"}, {"y": "string"}], "ret": "null", "do": [null]}]}]
}''')
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [Call("sort", [Ref("array"), FcnDef([{"x": AvroInt()}, {"y": AvroString()}], AvroNull(), [LiteralNull()])])],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"sort":["array",{"params": [{"x": "int"}, {"y": "string"}], "ret": "null", "do": [null]}]}]
}'''))

    def testRef(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [Ref("x")],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": ["x"]
}'''))

    def testNull(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [LiteralNull()],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [null]
}'''))

    def testBoolean(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [LiteralBoolean(True), LiteralBoolean(False)],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [true, false]
}'''))

    def testInt(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [LiteralInt(2)],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [2]
}'''))

    def testLong(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroLong(),
        AvroString(),
        [],
        [LiteralLong(2)],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "long",
  "output": "string",
  "action": [{"long": 2}]
}'''))
  
    def testFloat(self):
      self.assertEqual(
      EngineConfig(
      "test",
      Method.MAP,
      AvroInt(),
      AvroString(),
      [],
      [LiteralFloat(2.5)],
      [],
      {},
      None,
      None,
      {},
      {},
      None,
      None,
      None,
      {},
      {}),
      jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"float": 2.5}]
}'''))

    def testDouble(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [LiteralDouble(2.2)],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [2.2]
}'''))

    def testString(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [LiteralString("hello")],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"string": "hello"}]
}'''))
  
    def testBase64(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [LiteralBase64(bytesToString("hello".encode("utf-8")))],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"base64": "aGVsbG8="}]
}'''))

    def testLiteral(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [Literal(AvroRecord([AvroField("one", AvroInt()), AvroField("two", AvroDouble()), AvroField("three", AvroString())], "SimpleRecord"), '''{"one": 1, "two": 2.2, "three": "THREE"}''')],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"type":{"type":"record","name":"SimpleRecord","fields":[{"name":"one","type":"int"},{"name":"two","type":"double"},{"name":"three","type":"string"}]},"value":{"one":1,"two":2.2,"three":"THREE"}}]
}'''))
  
    def testNewRecord(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [NewObject({"one": LiteralInt(1), "two": LiteralDouble(2.2), "three": LiteralString("THREE")},
                       AvroRecord([AvroField("one", AvroInt()), AvroField("two", AvroDouble()), AvroField("three", AvroString())], "SimpleRecord"), AvroTypeBuilder())],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"new":{"one":{"int":1},"two":2.2,"three":{"string":"THREE"}},"type":{"type":"record","name":"SimpleRecord","fields":[{"name":"one","type":"int"},{"name":"two","type":"double"},{"name":"three","type":"string"}]}}]
}'''))
  
    def testNewArray(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [NewArray([LiteralInt(1), LiteralInt(2), LiteralInt(3)], AvroArray(AvroInt()), AvroTypeBuilder())],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"new":[1,2,3],"type":{"type":"array","items":"int"}}]
}'''))
  
    def testDo(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [Do([Ref("x"), Ref("y"), Ref("z")])],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"do":["x","y","z"]}]
}'''))
  
    def testLet(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [Let({"x": LiteralInt(3), "y": LiteralInt(4)})],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"let":{"x":3,"y":4}}]
}'''))
  
    def testSet(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [SetVar({"x": LiteralInt(3), "y": LiteralInt(4)})],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"set":{"x":3,"y":4}}]
}'''))
  
    def testAttrGet(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [AttrGet(Ref("c"), [Ref("a"), LiteralInt(1), LiteralString("b")])],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"attr":"c","path":["a",1,{"string":"b"}]}]
}'''))
  
    def testAttrSet(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [AttrTo(Ref("c"), [Ref("a"), LiteralInt(1), LiteralString("b")], LiteralDouble(2.2), None)],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"attr":"c","path":["a",1,{"string":"b"}],"to":2.2}]
}'''))
  
    def testCellGet(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [CellGet("c", [])],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"cell":"c"}]
}'''))
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [CellGet("c", [Ref("a"), LiteralInt(1), LiteralString("b")])],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"cell":"c","path":["a",1,{"string":"b"}]}]
}'''))
  
    def testCellSet(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [CellTo("c", [], LiteralDouble(2.2))],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"cell":"c","to":2.2}]
}'''))
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [CellTo("c", [Ref("a"), LiteralInt(1), LiteralString("b")], LiteralDouble(2.2))],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"cell":"c","path":["a",1,{"string":"b"}],"to":2.2}]
}'''))
  
    def testPoolGet(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [PoolGet("p", [Ref("a"), LiteralInt(1), LiteralString("b")])],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"pool":"p","path":["a",1,{"string":"b"}]}]
}'''))
  
    def testPoolSet(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [PoolTo("p", [Ref("a"), LiteralInt(1), LiteralString("b")], LiteralDouble(2.2), LiteralDouble(2.2))],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"pool":"p","path":["a",1,{"string":"b"}],"to":2.2,"init":2.2}]
}'''))

    def testIf(self):
        self.assertEqual(
        EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [If(LiteralBoolean(True), [Ref("x")], None), If(LiteralBoolean(True), [Ref("x")], [Ref("y")])],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}),
        jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"if":true,"then":["x"]}, {"if":true,"then":["x"],"else":["y"]}]
}'''))

    def testCond(self):
        self.assertEqual(
          EngineConfig(
          "test",
          Method.MAP,
          AvroInt(),
          AvroString(),
          [],
          [Cond([If(LiteralBoolean(False), [Ref("x")], None), If(LiteralBoolean(True), [Ref("y")], None)], None),
               Cond([If(LiteralBoolean(False), [Ref("x")], None), If(LiteralBoolean(True), [Ref("y")], None)], [Ref("z")])],
          [],
          {},
          None,
          None,
          {},
          {},
          None,
          None,
          None,
          {},
          {}),
          jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"cond":[{"if":false,"then":["x"]},{"if":true,"then":["y"]}]},
             {"cond":[{"if":false,"then":["x"]},{"if":true,"then":["y"]}],"else":["z"]}]
}'''))
  
    def testWhile(self):
        self.assertEqual(
          EngineConfig(
          "test",
          Method.MAP,
          AvroInt(),
          AvroString(),
          [],
          [While(LiteralBoolean(True), [Call("+", [LiteralInt(2), LiteralInt(2)])])],
          [],
          {},
          None,
          None,
          {},
          {},
          None,
          None,
          None,
          {},
          {}),
          jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"while":true,"do":[{"+":[2,2]}]}]
}'''))
  
    def testDoUntil(self):
        self.assertEqual(
          EngineConfig(
          "test",
          Method.MAP,
          AvroInt(),
          AvroString(),
          [],
          [DoUntil([Call("+", [LiteralInt(2), LiteralInt(2)])], LiteralBoolean(True))],
          [],
          {},
          None,
          None,
          {},
          {},
          None,
          None,
          None,
          {},
          {}),
          jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"do":[{"+":[2,2]}],"until":true}]
}'''))
  
    def testFor(self):
        self.assertEqual(
          EngineConfig(
          "test",
          Method.MAP,
          AvroInt(),
          AvroString(),
          [],
          [For({"i": LiteralInt(0)}, Call("<", [Ref("i"), LiteralInt(10)]), {"i": Call("+", [Ref("i"), LiteralInt(1)])}, [Ref("i")], False)],
          [],
          {},
          None,
          None,
          {},
          {},
          None,
          None,
          None,
          {},
          {}),
          jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"for":{"i":0},"while":{"<":["i",10]},"step":{"i":{"+":["i",1]}},"do":["i"]}]
}'''))
  
    def testForeach(self):
        self.assertEqual(
          EngineConfig(
          "test",
          Method.MAP,
          AvroInt(),
          AvroString(),
          [],
          [Foreach("x", Literal(AvroArray(AvroInt()), '''[1, 2, 3]'''), [Ref("x")], False)],
          [],
          {},
          None,
          None,
          {},
          {},
          None,
          None,
          None,
          {},
          {}),
          jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"foreach":"x","in":{"type":{"type":"array","items":"int"},"value":[1,2,3]},"do":["x"],"seq":false}]
}'''))
  
    def testForkeyval(self):
        self.assertEqual(
          EngineConfig(
          "test",
          Method.MAP,
          AvroInt(),
          AvroString(),
          [],
          [Forkeyval("k", "v", Literal(AvroMap(AvroInt()), '''{"one": 1, "two": 2, "three": 3}'''), [Ref("k")])],
          [],
          {},
          None,
          None,
          {},
          {},
          None,
          None,
          None,
          {},
          {}),
          jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"forkey":"k","forval":"v","in":{"type":{"type":"map","values":"int"},"value":{"one":1,"two":2,"three":3}},"do":["k"]}]
}'''))
  
    def testCast(self):
        self.assertEqual(
          EngineConfig(
          "test",
          Method.MAP,
          AvroInt(),
          AvroString(),
          [],
          [CastBlock(LiteralInt(3), [CastCase(AvroString(), "x", [Ref("x")]), CastCase(AvroInt(), "x", [Ref("x")])], True)],
          [],
          {},
          None,
          None,
          {},
          {},
          None,
          None,
          None,
          {},
          {}),
          jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"cast":{"int":3},"cases":[{"as":"string","named":"x","do":["x"]},{"as":"int","named":"x","do":["x"]}],"partial":true}]
}'''))
  
    def testDoc(self):
        self.assertEqual(
          EngineConfig(
          "test",
          Method.MAP,
          AvroInt(),
          AvroString(),
          [],
          [Doc("hello")],
          [],
          {},
          None,
          None,
          {},
          {},
          None,
          None,
          None,
          {},
          {}),
          jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"doc":"hello"}]
}'''))
  
    def testError(self):
        self.assertEqual(
          EngineConfig(
          "test",
          Method.MAP,
          AvroInt(),
          AvroString(),
          [],
          [Error("hello", None), Error("hello", -3)],
          [],
          {},
          None,
          None,
          {},
          {},
          None,
          None,
          None,
          {},
          {}),
          jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"error":"hello"}, {"error":"hello","code":-3}]
}'''))
  
    def testLog(self):
        self.assertEqual(
          EngineConfig(
          "test",
          Method.MAP,
          AvroInt(),
          AvroString(),
          [],
          [Log([LiteralString("hello")], None),
           Log([LiteralString("hello")], "DEBUG"),
           Log([Call("+", [LiteralInt(2), LiteralInt(2)])], None)],
          [],
          {},
          None,
          None,
          {},
          {},
          None,
          None,
          None,
          {},
          {}),
          jsonToAst('''{
  "name": "test",
  "input": "int",
  "output": "string",
  "action": [{"log":[{"string":"hello"}]},
             {"log":[{"string":"hello"}],"namespace":"DEBUG"},
             {"log":[{"+":[2,2]}]}]
}'''))
  
if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import *
from titus.util import callfcn, div
import titus.P as P
from titus.lib.array import argLowestN

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "model.cluster."

#################################################################### 

class Closest(LibFcn):
    name = prefix + "closest"
    sig = Sigs([
        Sig([{"datum": P.Array(P.Double())}, {"clusters": P.Array(P.WildRecord("C", {"center": P.Array(P.Double())}))}], P.Wildcard("C")),
        Sig([{"datum": P.Wildcard("A")}, {"clusters": P.Array(P.WildRecord("C", {"center": P.Wildcard("B")}))}, {"metric": P.Fcn([P.Wildcard("A"), P.Wildcard("B")], P.Double())}], P.Wildcard("C"))])
    errcodeBase = 29000
    def __call__(self, state, scope, pos, paramTypes, datum, clusters, *args):
        if len(clusters) == 0:
            raise PFARuntimeException("no clusters", self.errcodeBase + 0, self.name, pos)
        if len(args) == 1:
            metric, = args
            distances = [callfcn(state, scope, metric, [datum, x["center"]]) for x in clusters]
        else:
            distances = [sum((di - xi)**2 for di, xi in zip(datum, x["center"])) for x in clusters]
        index, = argLowestN(distances, 1, lambda a, b: a < b)
        return clusters[index]
provide(Closest())

class ClosestN(LibFcn):
    name = prefix + "closestN"
    sig = Sigs([
        Sig([{"n": P.Int()}, {"datum": P.Array(P.Double())}, {"clusters": P.Array(P.WildRecord("C", {"center": P.Array(P.Double())}))}], P.Array(P.Wildcard("C"))),
        Sig([{"n": P.Int()}, {"datum": P.Wildcard("A")}, {"clusters": P.Array(P.WildRecord("C", {"center": P.Wildcard("B")}))}, {"metric": P.Fcn([P.Wildcard("A"), P.Wildcard("B")], P.Double())}], P.Array(P.Wildcard("C")))])
    errcodeBase = 29010
    def __call__(self, state, scope, pos, paramTypes, n, datum, clusters, *args):
        if n < 0:
            raise PFARuntimeException("n must be nonnegative", self.errcodeBase + 0, self.name, pos)
        if len(args) == 1:
            metric, = args
            distances = [callfcn(state, scope, metric, [datum, x["center"]]) for x in clusters]
        else:
            distances = [sum((di - xi)**2 for di, xi in zip(datum, x["center"])) for x in clusters]
        indexes = argLowestN(distances, n, lambda a, b: a < b)
        return [clusters[i] for i in indexes]
provide(ClosestN())

class RandomSeeds(LibFcn):
    name = prefix + "randomSeeds"
    sig = Sig([{"data": P.Array(P.Array(P.Wildcard("A")))}, {"k": P.Int()}, {"newCluster": P.Fcn([P.Int(), P.Array(P.Wildcard("A"))], P.WildRecord("C", {"center": P.Array(P.Wildcard("B"))}))}], P.Array(P.Wildcard("C")))
    errcodeBase = 29020
    def __call__(self, state, scope, pos, paramTypes, data, k, newCluster):
        if k <= 0:
            raise PFARuntimeException("k must be greater than zero", self.errcodeBase + 0, self.name, pos)

        uniques = []
        for x in data:
            if x not in uniques:
                uniques.append(x)

        if len(uniques) < k:
            raise PFARuntimeException("not enough unique points", self.errcodeBase + 1, self.name, pos)

        sizes = set(len(x) for x in uniques)
        if len(sizes) != 1:
            raise PFARuntimeException("dimensions of vectors do not match", self.errcodeBase + 2, self.name, pos)

        state.rand.shuffle(uniques)
        selected = uniques[:k]
        
        return [callfcn(state, scope, newCluster, [i, list(vec)]) for i, vec in enumerate(selected)]
provide(RandomSeeds())

class KMeansIteration(LibFcn):
    name = prefix + "kmeansIteration"
    sig = Sig([{"data": P.Array(P.Array(P.Wildcard("A")))}, {"clusters": P.Array(P.WildRecord("C", {"center": P.Array(P.Wildcard("B"))}))}, {"metric": P.Fcn([P.Array(P.Wildcard("A")), P.Array(P.Wildcard("B"))], P.Double())}, {"update": P.Fcn([P.Array(P.Array(P.Wildcard("A"))), P.Wildcard("C")], P.Wildcard("C"))}], P.Array(P.Wildcard("C")))
    errcodeBase = 29030
    def __call__(self, state, scope, pos, paramTypes, data, clusters, metric, update):
        if len(data) == 0:
            raise PFARuntimeException("no data", self.errcodeBase + 0, self.name, pos)

        centers = [x["center"] for x in clusters]

        length = len(clusters)
        if length == 0:
            raise PFARuntimeException("no clusters", self.errcodeBase + 1, self.name, pos)

        matched = [[] for i in range(length)]

        for datum in data:
            besti = 0
            bestCenter = None
            bestDistance = 0.0
            i = 0
            while i < length:
                thisCenter = centers[i]
                thisDistance = callfcn(state, scope, metric, [datum, thisCenter])
                if bestCenter is None or thisDistance < bestDistance:
                    besti = i
                    bestCenter = thisCenter
                    bestDistance = thisDistance
                i += 1
            matched[besti].append(datum)

        out = []
        for i, matchedData in enumerate(matched):
            if len(matchedData) == 0:
                out.append(clusters[i])
            else:
                out.append(callfcn(state, scope, update, [matchedData, clusters[i]]))
        return out
provide(KMeansIteration())

class UpdateMean(LibFcn):
    name = prefix + "updateMean"
    sig = Sig([{"data": P.Array(P.Array(P.Double()))}, {"cluster": P.WildRecord("C", {"center": P.Array(P.Double())})}, {"weight": P.Double()}], P.Wildcard("C"))
    errcodeBase = 29040
    def __call__(self, state, scope, pos, paramTypes, data, cluster, weight):
        if len(data) == 0:
            raise PFARuntimeException("no data", self.errcodeBase + 0, self.name, pos)

        dimension = len(data[0])
        summ = [0.0 for i in range(dimension)]

        for vec in data:
            if len(vec) != dimension:
                raise PFARuntimeException("dimensions of vectors do not match", self.errcodeBase + 1, self.name, pos)
            for i in range(dimension):
                summ[i] += vec[i]

        vec = cluster["center"]
        if len(vec) != dimension:
            raise PFARuntimeException("dimensions of vectors do not match", self.errcodeBase + 1, self.name, pos)
        for i in range(dimension):
            summ[i] += weight * vec[i]

        denom = len(data) + weight
        for i in range(dimension):
            summ[i] = div(summ[i], denom)

        return dict(cluster, center=summ)
provide(UpdateMean())



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import io
import json

import avro.schema
from avro.io import BinaryEncoder, BinaryDecoder, DatumReader, DatumWriter

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.util import callfcn
from titus.errors import PFARuntimeException
import titus.P as P

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "map."

class ObjKey(object):
    def toKey(self, x, avroType):
        x = jsonEncoder(avroType, x, False)
        bytes = io.BytesIO()
        writer = DatumWriter(avroType.schema)
        writer.write(x, BinaryEncoder(bytes))
        bytes.flush()
        return (base64.b64encode(bytes.getvalue())).decode()

    def fromKey(self, key, avroType):
        bytes = io.BytesIO(base64.b64decode(key))
        reader = DatumReader(avroType.schema)
        return reader.read(BinaryDecoder(bytes))

def toLt(state, scope, lessThan):
    return lambda a, b: callfcn(state, scope, lessThan, [a, b])

#################################################################### basic access

class Len(LibFcn):
    name = prefix + "len"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}], P.Int())
    errcodeBase = 26000
    def __call__(self, state, scope, pos, paramTypes, m):
        return len(m)
provide(Len())

class Keys(LibFcn):
    name = prefix + "keys"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}], P.Array(P.String()))
    errcodeBase = 26010
    def __call__(self, state, scope, pos, paramTypes, m):
        return list(m.keys())
provide(Keys())

class Values(LibFcn):
    name = prefix + "values"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}], P.Array(P.Wildcard("A")))
    errcodeBase = 26020
    def __call__(self, state, scope, pos, paramTypes, m):
        return list(m.values())
provide(Values())

#################################################################### searching

class ContainsKey(LibFcn):
    name = prefix + "containsKey"
    sig = Sigs([Sig([{"m": P.Map(P.Wildcard("A"))}, {"key": P.String()}], P.Boolean()),
                Sig([{"m": P.Map(P.Wildcard("A"))}, {"fcn": P.Fcn([P.String()], P.Boolean())}], P.Boolean())])
    errcodeBase = 26030
    def __call__(self, state, scope, pos, paramTypes, m, key):
        if callable(key):
            for k in m:
                if callfcn(state, scope, key, [k]):
                    return True
            return False
        else:
            return key in m
provide(ContainsKey())

class ContainsValue(LibFcn):
    name = prefix + "containsValue"
    sig = Sigs([Sig([{"m": P.Map(P.Wildcard("A"))}, {"value": P.Wildcard("A")}], P.Boolean()),
                Sig([{"m": P.Map(P.Wildcard("A"))}, {"fcn": P.Fcn([P.Wildcard("A")], P.Boolean())}], P.Boolean())])
    errcodeBase = 26040
    def __call__(self, state, scope, pos, paramTypes, m, value):
        if callable(value):
            for v in list(m.values()):
                if callfcn(state, scope, value, [v]):
                    return True
            return False
        else:
            for v in list(m.values()):
                if v == value:
                    return True
            return False
provide(ContainsValue())

#################################################################### manipulation

class Add(LibFcn, ObjKey):
    name = prefix + "add"
    sig = Sigs([Sig([{"m": P.Map(P.Wildcard("A"))}, {"key": P.String()}, {"value": P.Wildcard("A")}], P.Map(P.Wildcard("A"))),
                Sig([{"m": P.Map(P.Wildcard("A"))}, {"item": P.Wildcard("A")}], P.Map(P.Wildcard("A")))])
    errcodeBase = 26050
    def __call__(self, state, scope, pos, paramTypes, m, *args):
        if len(args) == 2:
            key, value = args
            return dict(m, **{key: value})
        else:
            item, = args
            key = self.toKey(item, jsonNodeToAvroType(paramTypes[1]))
            return dict(m, **{key: item})
provide(Add())

class Remove(LibFcn):
    name = prefix + "remove"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}, {"key": P.String()}], P.Map(P.Wildcard("A")))
    errcodeBase = 26060
    def __call__(self, state, scope, pos, paramTypes, m, key):
        out = dict(m)
        try:
            del out[key]
        except KeyError:
            pass
        return out
provide(Remove())

class Only(LibFcn):
    name = prefix + "only"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}, {"keys": P.Array(P.String())}], P.Map(P.Wildcard("A")))
    errcodeBase = 26070
    def __call__(self, state, scope, pos, paramTypes, m, keys):
        return dict((k, v) for k, v in list(m.items()) if k in keys)
provide(Only())

class Except(LibFcn):
    name = prefix + "except"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}, {"keys": P.Array(P.String())}], P.Map(P.Wildcard("A")))
    errcodeBase = 26080
    def __call__(self, state, scope, pos, paramTypes, m, keys):
        return dict((k, v) for k, v in list(m.items()) if k not in keys)
provide(Except())

class Update(LibFcn):
    name = prefix + "update"
    sig = Sig([{"base": P.Map(P.Wildcard("A"))}, {"overlay": P.Map(P.Wildcard("A"))}], P.Map(P.Wildcard("A")))
    errcodeBase = 26090
    def __call__(self, state, scope, pos, paramTypes, base, overlay):
        return dict(list(base.items()) + list(overlay.items()))
provide(Update())

class Split(LibFcn):
    name = prefix + "split"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}], P.Array(P.Map(P.Wildcard("A"))))
    errcodeBase = 26100
    def __call__(self, state, scope, pos, paramTypes, m):
        return [{k: v} for k, v in list(m.items())]
provide(Split())

class Join(LibFcn):
    name = prefix + "join"
    sig = Sig([{"a": P.Array(P.Map(P.Wildcard("A")))}], P.Map(P.Wildcard("A")))
    errcodeBase = 26110
    def __call__(self, state, scope, pos, paramTypes, a):
        out = []
        for m in a:
            out.extend(list(m.items()))
        return dict(out)
provide(Join())    

#################################################################### min/max functions

def argHighestN(m, n, lt):
    out = []
    for k, x in list(m.items()):
        found = False
        for index, (bestk, bestx) in enumerate(out):
            if lt(bestx, x) or (not lt(x, bestx) and bestk > k):
                ind = index
                while ind < len(out) and not lt(out[ind][1], x) and not lt(x, out[ind][1]) and out[ind][0] < k:
                    ind += 1
                out.insert(ind, (k, x))
                found = True
                break
        if not found:
            out.append((k, x))
        if len(out) > n:
            out.pop()
    return [k for k, x in out]

def argLowestN(m, n, lt):
    out = []
    for k, x in list(m.items()):
        found = False
        for index, (bestk, bestx) in enumerate(out):
            if lt(x, bestx) or (not lt(bestx, x) and bestk > k):
                ind = index
                while ind < len(out) and not lt(out[ind][1], x) and not lt(x, out[ind][1]) and out[ind][0] < k:
                    ind += 1
                out.insert(ind, (k, x))
                found = True
                break
        if not found:
            out.append((k, x))
        if len(out) > n:
            out.pop()
    return [k for k, x in out]

class Argmax(LibFcn):
    name = prefix + "argmax"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}], P.String())
    errcodeBase = 26120
    def __call__(self, state, scope, pos, paramTypes, m):
        if len(m) == 0:
            raise PFARuntimeException("empty map", self.errcodeBase + 0, self.name, pos)
        else:
            return argHighestN(m, 1, lambda x, y: compare(jsonNodeToAvroType(paramTypes[0]).values, x, y) < 0)[0]
provide(Argmax())

class Argmin(LibFcn):
    name = prefix + "argmin"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}], P.String())
    errcodeBase = 26130
    def __call__(self, state, scope, pos, paramTypes, m):
        if len(m) == 0:
            raise PFARuntimeException("empty map", self.errcodeBase + 0, self.name, pos)
        else:
            return argLowestN(m, 1, lambda x, y: compare(jsonNodeToAvroType(paramTypes[0]).values, x, y) < 0)[0]
provide(Argmin())

class ArgmaxLT(LibFcn):
    name = prefix + "argmaxLT"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}, {"lessThan": P.Fcn([P.Wildcard("A"), P.Wildcard("A")], P.Boolean())}], P.String())
    errcodeBase = 26140
    def __call__(self, state, scope, pos, paramTypes, m, lessThan):
        if len(m) == 0:
            raise PFARuntimeException("empty map", self.errcodeBase + 0, self.name, pos)
        else:
            return argHighestN(m, 1, toLt(state, scope, lessThan))[0]
provide(ArgmaxLT())

class ArgminLT(LibFcn):
    name = prefix + "argminLT"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}, {"lessThan": P.Fcn([P.Wildcard("A"), P.Wildcard("A")], P.Boolean())}], P.String())
    errcodeBase = 26150
    def __call__(self, state, scope, pos, paramTypes, m, lessThan):
        if len(m) == 0:
            raise PFARuntimeException("empty map", self.errcodeBase + 0, self.name, pos)
        else:
            return argLowestN(m, 1, toLt(state, scope, lessThan))[0]
provide(ArgminLT())

class ArgmaxN(LibFcn):
    name = prefix + "argmaxN"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}, {"n": P.Int()}], P.Array(P.String()))
    errcodeBase = 26160
    def __call__(self, state, scope, pos, paramTypes, m, n):
        if len(m) == 0:
            raise PFARuntimeException("empty map", self.errcodeBase + 0, self.name, pos)
        elif n < 0:
            raise PFARuntimeException("n < 0", self.errcodeBase + 1, self.name, pos)
        else:
            return argHighestN(m, n, lambda x, y: compare(jsonNodeToAvroType(paramTypes[0]).values, x, y) < 0)
provide(ArgmaxN())

class ArgminN(LibFcn):
    name = prefix + "argminN"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}, {"n": P.Int()}], P.Array(P.String()))
    errcodeBase = 26170
    def __call__(self, state, scope, pos, paramTypes, m, n):
        if len(m) == 0:
            raise PFARuntimeException("empty map", self.errcodeBase + 0, self.name, pos)
        elif n < 0:
            raise PFARuntimeException("n < 0", self.errcodeBase + 1, self.name, pos)
        else:
            return argLowestN(m, n, lambda x, y: compare(jsonNodeToAvroType(paramTypes[0]).values, x, y) < 0)
provide(ArgminN())

class ArgmaxNLT(LibFcn):
    name = prefix + "argmaxNLT"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}, {"n": P.Int()}, {"lessThan": P.Fcn([P.Wildcard("A"), P.Wildcard("A")], P.Boolean())}], P.Array(P.String()))
    errcodeBase = 26180
    def __call__(self, state, scope, pos, paramTypes, m, n, lessThan):
        if len(m) == 0:
            raise PFARuntimeException("empty map", self.errcodeBase + 0, self.name, pos)
        elif n < 0:
            raise PFARuntimeException("n < 0", self.errcodeBase + 1, self.name, pos)
        else:
            return argHighestN(m, n, toLt(state, scope, lessThan))
provide(ArgmaxNLT())

class ArgminNLT(LibFcn):
    name = prefix + "argminNLT"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}, {"n": P.Int()}, {"lessThan": P.Fcn([P.Wildcard("A"), P.Wildcard("A")], P.Boolean())}], P.Array(P.String()))
    errcodeBase = 26190
    def __call__(self, state, scope, pos, paramTypes, m, n, lessThan):
        if len(m) == 0:
            raise PFARuntimeException("empty map", self.errcodeBase + 0, self.name, pos)
        elif n < 0:
            raise PFARuntimeException("n < 0", self.errcodeBase + 1, self.name, pos)
        else:
            return argLowestN(m, n, toLt(state, scope, lessThan))
provide(ArgminNLT())

#################################################################### set or set-like functions

class ToSet(LibFcn, ObjKey):
    name = prefix + "toset"
    sig = Sig([{"a": P.Array(P.Wildcard("A"))}], P.Map(P.Wildcard("A")))
    errcodeBase = 26200
    def __call__(self, state, scope, pos, paramTypes, a):
        return dict((self.toKey(x, jsonNodeToAvroType(paramTypes[0]["items"])), x) for x in a)
provide(ToSet())

class FromSet(LibFcn, ObjKey):
    name = prefix + "fromset"
    sig = Sig([{"s": P.Map(P.Wildcard("A"))}], P.Array(P.Wildcard("A")))
    errcodeBase = 26210
    def __call__(self, state, scope, pos, paramTypes, s):
        return list(s.values())
provide(FromSet())

class In(LibFcn, ObjKey):
    name = prefix + "in"
    sig = Sig([{"s": P.Map(P.Wildcard("A"))}, {"x": P.Wildcard("A")}], P.Boolean())
    errcodeBase = 26220
    def __call__(self, state, scope, pos, paramTypes, s, x):
        return self.toKey(x, jsonNodeToAvroType(paramTypes[0]["values"])) in s
provide(In())

class Union(LibFcn):
    name = prefix + "union"
    sig = Sig([{"a": P.Map(P.Wildcard("A"))}, {"b": P.Map(P.Wildcard("A"))}], P.Map(P.Wildcard("A")))
    errcodeBase = 26230
    def __call__(self, state, scope, pos, paramTypes, a, b):
        return dict(list(a.items()) + list(b.items()))
provide(Union())

class Intersection(LibFcn):
    name = prefix + "intersection"
    sig = Sig([{"a": P.Map(P.Wildcard("A"))}, {"b": P.Map(P.Wildcard("A"))}], P.Map(P.Wildcard("A")))
    errcodeBase = 26240
    def __call__(self, state, scope, pos, paramTypes, a, b):
        return dict((k, v) for k, v in list(a.items()) if k in b)
provide(Intersection())

class Diff(LibFcn):
    name = prefix + "diff"
    sig = Sig([{"a": P.Map(P.Wildcard("A"))}, {"b": P.Map(P.Wildcard("A"))}], P.Map(P.Wildcard("A")))
    errcodeBase = 26250
    def __call__(self, state, scope, pos, paramTypes, a, b):
        return dict((k, v) for k, v in list(a.items()) if k not in b)
provide(Diff())

class SymDiff(LibFcn):
    name = prefix + "symdiff"
    sig = Sig([{"a": P.Map(P.Wildcard("A"))}, {"b": P.Map(P.Wildcard("A"))}], P.Map(P.Wildcard("A")))
    errcodeBase = 26260
    def __call__(self, state, scope, pos, paramTypes, a, b):
        cc = set(a.keys()).symmetric_difference(set(b.keys()))
        return dict((k, v) for k, v in list(a.items()) + list(b.items()) if k in cc)
provide(SymDiff())

class Subset(LibFcn):
    name = prefix + "subset"
    sig = Sig([{"little": P.Map(P.Wildcard("A"))}, {"big": P.Map(P.Wildcard("A"))}], P.Boolean())
    errcodeBase = 26270
    def __call__(self, state, scope, pos, paramTypes, little, big):
        return all(k in big for k in list(little.keys()))
provide(Subset())

class Disjoint(LibFcn):
    name = prefix + "disjoint"
    sig = Sig([{"a": P.Map(P.Wildcard("A"))}, {"b": P.Map(P.Wildcard("A"))}], P.Boolean())
    errcodeBase = 26280
    def __call__(self, state, scope, pos, paramTypes, a, b):
        return len(set(a.keys()).intersection(set(b.keys()))) == 0
provide(Disjoint())

#################################################################### functional programming

class MapApply(LibFcn):
    name = prefix + "map"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}, {"fcn": P.Fcn([P.Wildcard("A")], P.Wildcard("B"))}], P.Map(P.Wildcard("B")))
    errcodeBase = 26290
    def __call__(self, state, scope, pos, paramTypes, m, fcn):
        return dict((k, callfcn(state, scope, fcn, [v])) for k, v in list(m.items()))
provide(MapApply())

class MapWithKey(LibFcn):
    name = prefix + "mapWithKey"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}, {"fcn": P.Fcn([P.String(), P.Wildcard("A")], P.Wildcard("B"))}], P.Map(P.Wildcard("B")))
    errcodeBase = 26300
    def __call__(self, state, scope, pos, paramTypes, m, fcn):
        return dict((k, callfcn(state, scope, fcn, [k, v])) for k, v in list(m.items()))
provide(MapWithKey())

class Filter(LibFcn):
    name = prefix + "filter"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}, {"fcn": P.Fcn([P.Wildcard("A")], P.Boolean())}], P.Map(P.Wildcard("A")))
    errcodeBase = 26310
    def __call__(self, state, scope, pos, paramTypes, m, fcn):
        return dict((k, v) for k, v in list(m.items()) if callfcn(state, scope, fcn, [v]))
provide(Filter())

class FilterWithKey(LibFcn):
    name = prefix + "filterWithKey"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}, {"fcn": P.Fcn([P.String(), P.Wildcard("A")], P.Boolean())}], P.Map(P.Wildcard("A")))
    errcodeBase = 26320
    def __call__(self, state, scope, pos, paramTypes, m, fcn):
        return dict((k, v) for k, v in list(m.items()) if callfcn(state, scope, fcn, [k, v]))
provide(FilterWithKey())

class FilterMap(LibFcn):
    name = prefix + "filterMap"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}, {"fcn": P.Fcn([P.Wildcard("A")], P.Union([P.Wildcard("B"), P.Null()]))}], P.Map(P.Wildcard("B")))
    errcodeBase = 26330
    def __call__(self, state, scope, pos, paramTypes, m, fcn):
        typeNames = [jsonNodeToAvroType(t).name for t in paramTypes[1]["ret"] if t != "null"]
        out = {}
        for k, v in list(m.items()):
            vv = callfcn(state, scope, fcn, [v])
            if vv is not None:
                if isinstance(vv, dict) and len(vv) == 1 and list(vv.keys())[0] in typeNames:
                    tag, value = list(vv.items())[0]
                else:
                    value = vv
                out[k] = value
        return out
provide(FilterMap())

class FilterMapWithKey(LibFcn):
    name = prefix + "filterMapWithKey"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}, {"fcn": P.Fcn([P.String(), P.Wildcard("A")], P.Union([P.Wildcard("B"), P.Null()]))}], P.Map(P.Wildcard("B")))
    errcodeBase = 26340
    def __call__(self, state, scope, pos, paramTypes, m, fcn):
        typeNames = [jsonNodeToAvroType(t).name for t in paramTypes[1]["ret"] if t != "null"]
        out = {}
        for k, v in list(m.items()):
            vv = callfcn(state, scope, fcn, [k, v])
            if vv is not None:
                if isinstance(vv, dict) and len(vv) == 1 and list(vv.keys())[0] in typeNames:
                    tag, value = list(vv.items())[0]
                else:
                    value = vv
                out[k] = value
        return out
provide(FilterMapWithKey())

class FlatMap(LibFcn):
    name = prefix + "flatMap"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}, {"fcn": P.Fcn([P.Wildcard("A")], P.Map(P.Wildcard("B")))}], P.Map(P.Wildcard("B")))
    errcodeBase = 26350
    def __call__(self, state, scope, pos, paramTypes, m, fcn):
        out = {}
        for key, value in list(m.items()):
            for k, v in list(callfcn(state, scope, fcn, [value]).items()):
                out[k] = v
        return out
provide(FlatMap())

class FlatMapWithKey(LibFcn):
    name = prefix + "flatMapWithKey"
    sig = Sig([{"m": P.Map(P.Wildcard("A"))}, {"fcn": P.Fcn([P.String(), P.Wildcard("A")], P.Map(P.Wildcard("B")))}], P.Map(P.Wildcard("B")))
    errcodeBase = 26360
    def __call__(self, state, scope, pos, paramTypes, m, fcn):
        out = {}
        for key, value in list(m.items()):
            for k, v in list(callfcn(state, scope, fcn, [key, value]).items()):
                out[k] = v
        return out
provide(FlatMapWithKey())

class ZipMap(LibFcn):
    name = prefix + "zipmap"
    sig = Sigs([Sig([{"a": P.Map(P.Wildcard("A"))}, {"b": P.Map(P.Wildcard("B"))}, {"fcn": P.Fcn([P.Wildcard("A"), P.Wildcard("B")], P.Wildcard("Z"))}], P.Map(P.Wildcard("Z"))),
                Sig([{"a": P.Map(P.Wildcard("A"))}, {"b": P.Map(P.Wildcard("B"))}, {"c": P.Map(P.Wildcard("C"))}, {"fcn": P.Fcn([P.Wildcard("A"), P.Wildcard("B"), P.Wildcard("C")], P.Wildcard("Z"))}], P.Map(P.Wildcard("Z"))),
                Sig([{"a": P.Map(P.Wildcard("A"))}, {"b": P.Map(P.Wildcard("B"))}, {"c": P.Map(P.Wildcard("C"))}, {"d": P.Map(P.Wildcard("D"))}, {"fcn": P.Fcn([P.Wildcard("A"), P.Wildcard("B"), P.Wildcard("C"), P.Wildcard("D")], P.Wildcard("Z"))}], P.Map(P.Wildcard("Z")))])
    errcodeBase = 26370
    def __call__(self, state, scope, pos, paramTypes, *args):
        fcn = args[-1]
        maps = args[:-1]
        keys = list(maps[0].keys())
        for m in maps:
            if keys != list(m.keys()):
                raise PFARuntimeException("misaligned maps", self.errcodeBase + 0, self.name, pos)
        out = {}
        for k in keys:
            out[k] = callfcn(state, scope, fcn, [x[k] for x in maps])
        return out
provide(ZipMap())

class ZipMapWithKey(LibFcn):
    name = prefix + "zipmapWithKey"
    sig = Sigs([Sig([{"a": P.Map(P.Wildcard("A"))}, {"b": P.Map(P.Wildcard("B"))}, {"fcn": P.Fcn([P.String(), P.Wildcard("A"), P.Wildcard("B")], P.Wildcard("Z"))}], P.Map(P.Wildcard("Z"))),
                Sig([{"a": P.Map(P.Wildcard("A"))}, {"b": P.Map(P.Wildcard("B"))}, {"c": P.Map(P.Wildcard("C"))}, {"fcn": P.Fcn([P.String(), P.Wildcard("A"), P.Wildcard("B"), P.Wildcard("C")], P.Wildcard("Z"))}], P.Map(P.Wildcard("Z"))),
                Sig([{"a": P.Map(P.Wildcard("A"))}, {"b": P.Map(P.Wildcard("B"))}, {"c": P.Map(P.Wildcard("C"))}, {"d": P.Map(P.Wildcard("D"))}, {"fcn": P.Fcn([P.String(), P.Wildcard("A"), P.Wildcard("B"), P.Wildcard("C"), P.Wildcard("D")], P.Wildcard("Z"))}], P.Map(P.Wildcard("Z")))])
    errcodeBase = 26380
    def __call__(self, state, scope, pos, paramTypes, *args):
        fcn = args[-1]
        maps = args[:-1]
        keys = list(maps[0].keys())
        for m in maps:
            if keys != list(m.keys()):
                raise PFARuntimeException("misaligned maps", self.errcodeBase + 0, self.name, pos)
        out = {}
        for k in keys:
            out[k] = callfcn(state, scope, fcn, [k] + [x[k] for x in maps])
        return out
provide(ZipMapWithKey())

#################################################################### functional tests

class Corresponds(LibFcn):
    name = prefix + "corresponds"
    sig = Sig([{"a": P.Map(P.Wildcard("A"))}, {"b": P.Map(P.Wildcard("B"))}, {"fcn": P.Fcn([P.Wildcard("A"), P.Wildcard("B")], P.Boolean())}], P.Boolean())
    errcodeBase = 26390
    def __call__(self, state, scope, pos, paramTypes, a, b, fcn):
        aset = set(a.keys())
        bset = set(b.keys())
        if aset != bset:
            return False
        else:
            return all(callfcn(state, scope, fcn, [a[k], b[k]]) for k in aset)
provide(Corresponds())

class CorrespondsWithKey(LibFcn):
    name = prefix + "correspondsWithKey"
    sig = Sig([{"a": P.Map(P.Wildcard("A"))}, {"b": P.Map(P.Wildcard("B"))}, {"fcn": P.Fcn([P.String(), P.Wildcard("A"), P.Wildcard("B")], P.Boolean())}], P.Boolean())
    errcodeBase = 26400
    def __call__(self, state, scope, pos, paramTypes, a, b, fcn):
        aset = set(a.keys())
        bset = set(b.keys())
        if aset != bset:
            return False
        else:
            return all(callfcn(state, scope, fcn, [k, a[k], b[k]]) for k in aset)
provide(CorrespondsWithKey())



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from titus.errors import PFAInitializationException

class EngineOptions(object):
    """Represents the implementation options of a PFA document at runtime (the ``options`` section)."""

    def __init__(self, requestedOptions, hostOptions):
        """:type requestedOptions: dict from string to Pythonized JSON
        :param requestedOptions: options explicitly requested in the PFA document
        :type hostOptions: dict from string to Pythonized JSON
        :param hostOptions: options overridden by host environment
        """

        combinedOptions = {} if requestedOptions is None else dict(requestedOptions)
        if hostOptions is not None:
            combinedOptions.update(hostOptions)

        def longOpt(name, default):
            out = combinedOptions.get(name, default)
            try:
                return int(out)
            except ValueError:
                raise PFAInitializationException(name + " must be an integral number")

        self.timeout = longOpt("timeout", -1)
        self.timeout_begin = longOpt("timeout", self.timeout)
        self.timeout_action = longOpt("timeout", self.timeout)
        self.timeout_end = longOpt("timeout", self.timeout)

        # ...



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class AvroException(RuntimeError):
    """Exception for errors in deserializing or serializing Avro data."""
    pass

class SchemaParseException(RuntimeError):
    """Exception for errors in parsing an Avro schema."""
    pass

class PFAException(RuntimeError):
    """Base trait for any exceptions from PFA."""
    pass

class PrettyPfaException(PFAException):
    """Exceptions found in converting PrettyPFA into PFA."""
    pass

class PFASyntaxException(PFAException):
    """Exception for PFA syntax errors (JSON does not describe a PFA scoring engine)."""
    def __init__(self, message, pos):
        """:type message: string
        :param message: syntax error message
        :type pos: string or ``None``
        :param pos: source file position
        """
        self.pos = pos
        if pos is None or pos == "":
            super(PFASyntaxException, self).__init__("PFA syntax error: " + message)
        else:
            super(PFASyntaxException, self).__init__("PFA syntax error " + pos + ": " + message)

class PFASemanticException(PFAException):
    """Exception for PFA semantic errors (higher-level PFA rules are broken, such as type mismatches)."""
    def __init__(self, message, pos):
        """:type message: string
        :param message: semantic error message
        :type pos: string or ``None``
        :param pos: source file position
        """
        self.pos = pos
        if pos is None or pos == "":
            super(PFASemanticException, self).__init__("PFA semantic error: " + message)
        else:
            super(PFASemanticException, self).__init__("PFA semantic error " + pos + ": " + message)

class PFAInitializationException(PFAException):
    """Exception for PFA errors encountered during initialization (when cells/pools are being interpreted, almost ready to start evaluating)."""
    def __init__(self, message):
        """:type message: string
        :param message: initialization error message
        """
        super(PFAInitializationException, self).__init__("PFA initialization error: " + message)

class PFARuntimeException(PFAException):
    """Exception encountered at runtime in a PFA library call."""
    def __init__(self, message, code, fcnName, pos):
        """:type message: string
        :param message: library function error message
        """
        self.message = message
        self.code = code
        self.fcnName = fcnName
        self.pos = pos
        super(PFARuntimeException, self).__init__("{0} in {1} (#{2})".format(message, fcnName, code) + ("" if pos is None else " " + pos))

class PFAUserException(PFAException):
    """Exception deliberately invoked by the PFA author in a ``{"error": "xxx"}``` special form."""
    def __init__(self, message, code, pos):
        """:type message: string
        :param message: user-defined error message
        :type code: integer
        :param code: user-defined error code
        """
        self.message = message
        self.code = code
        self.pos = pos
        super(PFAUserException, self).__init__(message + ("" if code is None else " ({0})".format(code)) + ("" if pos is None else " " + pos))

class PFATimeoutException(PFAException):
    """Exception encountered at runtime from a PFA begin, action, end, or merge process taking too long (possible infinite loop)."""
    def __init__(self, message):
        """:type message: string
        :param message: timeout error message
        """
        super(PFATimeoutException, self).__init__("PFA timeout error: " + message)
